#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Module provides a Trigger class and related tools to create triggers in the database via the ORM.

It supports SQLite, MySQL and PostgreSQL with dialect-specific SQL generation.
"""

from abc import ABC, abstractmethod
from enum import StrEnum
from typing import Any, Self, cast

import peewee
from peewee import Model

from mafw.db.db_types import PeeweeModelWithMeta
from mafw.mafw_errors import MissingSQLStatement, UnsupportedDatabaseError
from mafw.tools.regexp import normalize_sql_spaces


def and_(*conditions: str) -> str:
    """
    Concatenates conditions with logical AND.

    :param conditions: The condition to join.
    :type conditions: str
    :return: The and-concatenated string of conditions
    :rtype: str
    """
    conditions_l = [f'({c})' for c in conditions]
    return ' AND '.join(conditions_l)


def or_(*conditions: str) -> str:
    """
    Concatenates conditions with logical OR.

    :param conditions: The condition to join.
    :type conditions: str
    :return: The or-concatenated string of conditions.
    :rtype: str
    """
    conditions_l = [f'({c})' for c in conditions]
    return ' OR '.join(conditions_l)


class TriggerWhen(StrEnum):
    """String enumerator for the trigger execution time (Before, After or Instead Of)"""

    Before = 'BEFORE'
    After = 'AFTER'
    Instead = 'INSTEAD OF'


class TriggerAction(StrEnum):
    """String enumerator for the trigger action (Delete, Insert, Update)"""

    Delete = 'DELETE'
    Insert = 'INSERT'
    Update = 'UPDATE'


class TriggerDialect(ABC):
    """Abstract base class for database-specific trigger SQL generation."""

    @abstractmethod
    def create_trigger_sql(self, trigger: 'Trigger') -> str:
        """
        Generate the SQL to create a trigger for a specific database dialect.

        :param trigger: The trigger object
        :return: SQL string to create the trigger
        """
        pass  # pragma: no cover

    @abstractmethod
    def drop_trigger_sql(self, trigger_name: str, safe: bool = True) -> str:
        """
        Generate the SQL to drop a trigger for a specific database dialect.

        :param trigger_name: The name of the trigger to drop
        :param safe: If True, add an IF EXISTS clause. Defaults to True.
        :return: SQL string to drop the trigger
        """
        pass  # pragma: no cover

    @abstractmethod
    def supports_trigger_type(self, when: TriggerWhen, action: TriggerAction, on_view: bool = False) -> bool:
        """
        Check if the database supports the specified trigger type.

        :param when: When the trigger should fire (BEFORE, AFTER, INSTEAD OF)
        :param action: The action that triggers the trigger (INSERT, UPDATE, DELETE)
        :param on_view: Whether the trigger is on a view
        :return: True if supported, False otherwise
        """
        pass  # pragma: no cover

    @abstractmethod
    def supports_safe_create(self) -> bool:
        """
        Check if the database supports IF NOT EXISTS for triggers.

        :return: True if supported, False otherwise
        """
        pass  # pragma: no cover

    @abstractmethod
    def supports_update_of_columns(self) -> bool:
        """
        Check if the database supports column-specific UPDATE triggers.

        :return: True if supported, False otherwise
        """
        pass  # pragma: no cover

    @abstractmethod
    def supports_when_clause(self) -> bool:
        """
        Check if the database supports WHEN conditions.

        :return: True if supported, False otherwise
        """
        pass  # pragma: no cover


class SQLiteDialect(TriggerDialect):
    """SQLite-specific trigger SQL generation."""

    def create_trigger_sql(self, trigger: 'Trigger') -> str:
        """Generate SQLite trigger SQL."""
        if_not_exists = 'IF NOT EXISTS' if trigger.safe else ''
        of_columns = (
            f'OF {", ".join(trigger.update_columns)}'
            if trigger.trigger_action == TriggerAction.Update and trigger.update_columns
            else ''
        )
        for_each_row = 'FOR EACH ROW' if trigger.for_each_row else ''
        when_clause = f'WHEN {" AND ".join(trigger._when_list)}' if trigger._when_list else ''
        sql_statements = '\n'.join(trigger._sql_list)

        return normalize_sql_spaces(
            f'CREATE TRIGGER {if_not_exists} {trigger.trigger_name}\n'
            f'{trigger.trigger_when} {trigger.trigger_action} {of_columns} ON {trigger.target_table}\n'
            f'{for_each_row} {when_clause}\n'
            f'BEGIN\n'
            f'{sql_statements}\n'
            f'END;'
        )

    def drop_trigger_sql(self, trigger_name: str, safe: bool = True) -> str:
        """Generate SQLite drop trigger SQL."""
        return normalize_sql_spaces(f'DROP TRIGGER {"IF EXISTS" if safe else ""} {trigger_name}')

    def supports_trigger_type(self, when: TriggerWhen, action: TriggerAction, on_view: bool = False) -> bool:
        """SQLite supports all trigger types except INSTEAD OF on tables (only on views)."""
        if when == TriggerWhen.Instead and not on_view:
            return False
        return True

    def supports_safe_create(self) -> bool:
        """SQLite supports IF NOT EXISTS for triggers."""
        return True

    def supports_update_of_columns(self) -> bool:
        """SQLite supports column-specific UPDATE triggers."""
        return True

    def supports_when_clause(self) -> bool:
        """SQLite supports WHEN conditions."""
        return True


class MySQLDialect(TriggerDialect):
    """MySQL-specific trigger SQL generation."""

    def create_trigger_sql(self, trigger: 'Trigger') -> str:
        """Generate MySQL trigger SQL."""
        # MySQL doesn't support INSTEAD OF triggers
        # MySQL doesn't support column-specific UPDATE triggers
        # MySQL requires FOR EACH ROW

        if_not_exists = 'IF NOT EXISTS' if trigger.safe else ''

        # In MySQL, we need to convert WHEN conditions to IF/THEN/END IF blocks
        sql_statements = []

        # If there are conditional statements, wrap them in IF blocks
        if trigger._when_list:
            condition = ' AND '.join(trigger._when_list)
            # Start the IF block
            sql_statements.append(f'IF {condition} THEN')

            # Add the SQL statements with indentation
            for stmt in trigger._sql_list:
                sql_statements.append(f'  {stmt}')

            # Close the IF block
            sql_statements.append('END IF;')
        else:
            # No conditions, just add the SQL statements directly
            sql_statements.extend(trigger._sql_list)

        # Join all statements
        trigger_body = '\n'.join(sql_statements)

        # Construct the final SQL
        sql = (
            f'CREATE TRIGGER {if_not_exists} {trigger.trigger_name}\n'
            f'{trigger.trigger_when} {trigger.trigger_action} ON {trigger.target_table}\n'
            f'FOR EACH ROW\n'
            f'BEGIN\n'
            f'{trigger_body}\n'
            f'END;'
        )
        return normalize_sql_spaces(sql)

    def drop_trigger_sql(self, trigger_name: str, safe: bool = True) -> str:
        """Generate MySQL drop trigger SQL."""
        return normalize_sql_spaces(f'DROP TRIGGER {"IF EXISTS" if safe else ""} {trigger_name}')

    def supports_trigger_type(self, when: TriggerWhen, action: TriggerAction, on_view: bool = False) -> bool:
        """MySQL doesn't support INSTEAD OF triggers."""
        return when != TriggerWhen.Instead

    def supports_safe_create(self) -> bool:
        """MySQL supports IF NOT EXISTS for triggers."""
        return True

    def supports_update_of_columns(self) -> bool:
        """MySQL doesn't support column-specific UPDATE triggers."""
        return False

    def supports_when_clause(self) -> bool:
        """MySQL supports conditions but through WHERE instead of WHEN."""
        return True


class PostgreSQLDialect(TriggerDialect):
    """PostgreSQL-specific trigger SQL generation."""

    def create_trigger_sql(self, trigger: 'Trigger') -> str:
        """Generate PostgreSQL trigger SQL."""
        # PostgreSQL handles INSTEAD OF differently
        # PostgreSQL uses functions for trigger bodies

        function_name = f'fn_{trigger.trigger_name}'

        # First create the function
        function_sql = f'CREATE OR REPLACE FUNCTION {function_name}() RETURNS TRIGGER AS $$\nBEGIN\n'

        # Add WHEN condition as IF statements if needed
        if trigger._when_list:
            when_condition = ' AND '.join(trigger._when_list)
            function_sql += f'  IF {when_condition} THEN\n'
            # Indent SQL statements
            sql_statements = '\n'.join(['    ' + self._clean_sql(sql) for sql in trigger._sql_list])
            function_sql += f'{sql_statements}\n  END IF;\n'
        else:
            # Indent SQL statements
            sql_statements = '\n'.join(['    ' + self._clean_sql(sql) for sql in trigger._sql_list])
            function_sql += f'{sql_statements}\n'

        # For AFTER triggers, we need to return NULL or NEW
        if trigger.trigger_when == TriggerWhen.After:
            function_sql += '  RETURN NULL;\n'
        # For BEFORE or INSTEAD OF triggers, we need to return NEW
        else:
            function_sql += '  RETURN NEW;\n'

        function_sql += 'END;\n$$ LANGUAGE plpgsql;'

        # Then create the trigger - PostgreSQL doesn't support IF NOT EXISTS for triggers before v14
        # We'll handle this through a conditional drop
        drop_if_exists = ''
        if trigger.safe:
            drop_if_exists = f'DROP TRIGGER IF EXISTS {trigger.trigger_name} ON {trigger.target_table} CASCADE;\n'

        # PostgreSQL uses different syntax for INSTEAD OF (only allowed on views)
        trigger_when = trigger.trigger_when

        # Column-specific triggers in PostgreSQL
        of_columns = (
            f'OF {", ".join(trigger.update_columns)}'
            if trigger.update_columns and trigger.trigger_action == TriggerAction.Update
            else ''
        )

        for_each = 'FOR EACH ROW' if trigger.for_each_row else 'FOR EACH STATEMENT'

        trigger_sql = (
            f'{drop_if_exists}'
            f'CREATE TRIGGER  {trigger.trigger_name}\n'
            f'{trigger_when} {trigger.trigger_action} {of_columns} ON {trigger.target_table}\n'
            f'{for_each}\n'
            f'EXECUTE FUNCTION {function_name}();'
        )

        sql = f'{function_sql}\n\n{trigger_sql}'

        return normalize_sql_spaces(sql)

    def _clean_sql(self, sql: str) -> str:
        """
        Remove RETURNING clauses from SQL statements for PostgreSQL trigger functions.

        :param sql: The SQL statement
        :return: SQL statement without RETURNING clause
        """
        # Find the RETURNING clause position - case insensitive search
        sql_upper = sql.upper()
        returning_pos = sql_upper.find('RETURNING')

        # If RETURNING exists, remove it and everything after it up to the semicolon
        if returning_pos != -1:
            semicolon_pos = sql.find(';', returning_pos)
            if semicolon_pos != -1:
                return sql[:returning_pos] + ';'
            return sql[:returning_pos]
        return sql

    def drop_trigger_sql(self, trigger_name: str, safe: bool = True) -> str:
        """Generate PostgreSQL drop trigger SQL."""
        function_name = f'fn_{trigger_name}'
        return normalize_sql_spaces(
            f'DROP TRIGGER {"IF EXISTS" if safe else ""} {trigger_name} ON table_name;\n'
            f'DROP FUNCTION {"IF EXISTS" if safe else ""} {function_name}();'
        )

    def supports_trigger_type(self, when: TriggerWhen, action: TriggerAction, on_view: bool = False) -> bool:
        """PostgreSQL supports INSTEAD OF only on views."""
        if when == TriggerWhen.Instead and not on_view:
            return False
        return True

    def supports_safe_create(self) -> bool:
        """PostgreSQL doesn't support IF NOT EXISTS for triggers before v14, but we implement safety differently."""
        return True  # We report True but handle it with DROP IF EXISTS

    def supports_update_of_columns(self) -> bool:
        """PostgreSQL supports column-specific UPDATE triggers."""
        return True

    def supports_when_clause(self) -> bool:
        """PostgreSQL supports WHEN conditions."""
        return True


class Trigger:
    """Trigger template wrapper for use with peewee ORM."""

    # noinspection PyProtectedMember
    def __init__(
        self,
        trigger_name: str,
        trigger_type: tuple[TriggerWhen, TriggerAction],
        source_table: type[Model] | Model | str,
        safe: bool = False,
        for_each_row: bool = False,
        update_columns: list[str] | None = None,
        on_view: bool = False,  # Added parameter to indicate if the target is a view
    ):
        """
        Constructor parameters:

        :param trigger_name: The name of this trigger. It needs to be unique!
        :type trigger_name: str
        :param trigger_type: A tuple with :class:`TriggerWhen` and :class:`TriggerAction` to specify on which action
            the trigger should be invoked and if before, after or instead of.
        :type trigger_type: tuple[TriggerWhen, TriggerAction]
        :param source_table: The table originating the trigger. It can be a model class, instance, or also the name of
            the table.
        :type source_table: type[Model] | Model | str
        :param safe: A boolean flag to define if in the trigger creation statement a 'IF NOT EXISTS' clause should be
            included. Defaults to False
        :type safe: bool, Optional
        :param for_each_row: A boolean flag to repeat the script content for each modified row in the table.
            Defaults to False.
        :type for_each_row: bool, Optional
        :param update_columns: A list of column names. When defining a trigger on a table update, it is possible to
            restrict the firing of the trigger to the cases when a subset of all columns have been updated. An column
            is updated also when the new value is equal to the old one. If you want to discriminate this case, use the
            :meth:`add_when` method. Defaults to None.
        :type update_columns: list[str], Optional
        :param on_view: A boolean flag to indicate if the target is a view. This affects the support for INSTEAD OF.
            Defaults to False.
        :type on_view: bool, Optional
        """
        self.trigger_name = trigger_name
        self.trigger_type = trigger_type
        self._trigger_when, self._trigger_op = self.trigger_type
        self.update_columns = update_columns or []
        self.on_view = on_view

        if isinstance(source_table, type):
            model_cls = cast(PeeweeModelWithMeta, source_table)
            self.target_table = model_cls._meta.table_name
        elif isinstance(source_table, Model):
            model_instance = cast(PeeweeModelWithMeta, source_table)
            self.target_table = model_instance._meta.table_name
        else:
            self.target_table = source_table

        self.safe = safe
        self.for_each_row = for_each_row

        self._when_list: list[str] = []
        self._sql_list: list[str] = []
        self._database: peewee.Database | peewee.DatabaseProxy | None = None
        self._dialect: TriggerDialect | None = None

    @property
    def trigger_action(self) -> TriggerAction:
        return self._trigger_op

    @trigger_action.setter
    def trigger_action(self, action: TriggerAction) -> None:
        self._trigger_op = action

    @property
    def trigger_when(self) -> TriggerWhen:
        return self._trigger_when

    @trigger_when.setter
    def trigger_when(self, when: TriggerWhen) -> None:
        self._trigger_when = when

    def __setattr__(self, key: Any, value: Any) -> None:
        if key == 'safe':
            self.if_not_exists = 'IF NOT EXISTS' if value else ''
        elif key == 'for_each_row':
            self._for_each_row = 'FOR EACH ROW' if value else ''
        else:
            super().__setattr__(key, value)

    def __getattr__(self, item: str) -> Any:
        """
        Custom attribute getter for computed properties.

        :param item: The attribute name to get
        :return: The attribute value
        :raises AttributeError: If the attribute doesn't exist
        """
        if item == 'safe':
            return hasattr(self, 'if_not_exists') and self.if_not_exists == 'IF NOT EXISTS'
        elif item == 'for_each_row':
            return hasattr(self, '_for_each_row') and self._for_each_row == 'FOR EACH ROW'
        else:
            # Raise AttributeError for non-existent attributes (standard Python behavior)
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}'")

    def add_sql(self, sql: str | peewee.Query) -> Self:
        """
        Add an SQL statement to be executed by the trigger.

        The ``sql`` can be either a string containing the sql statement, or it can be any other peewee Query.

        For example:

        .. code-block:: python

            # assuming you have created a trigger ...

            sql = AnotherTable.insert(
                field1=some_value, field2=another_value
            )
            trigger.add_sql(sql)

        In this way the SQL code is generated with parametric placeholder if needed.

        :param sql: The SQL statement.
        :type sql: str | peewee.Query
        :return: self for easy chaining
        :rtype: Trigger
        """
        if not isinstance(sql, str):
            sql = str(sql)
        sql = sql.strip()
        sql = chr(9) + sql
        if not sql.endswith(';'):
            sql += ';'
        self._sql_list.append(sql)
        return self

    def add_when(self, *conditions: str) -> Self:
        """
        Add conditions to the `when` statements.

        Conditions are logically ANDed.
        To have mixed `OR` and `AND` logic, use the functions :func:`and_` and :func:`or_`.

        :param conditions: Conditions to be added with logical AND
        :type conditions: str
        :return: self for easy chaining
        :rtype: Trigger
        """
        conditions_l = [f'({c.strip()})' for c in conditions]
        self._when_list.append(f'({" AND ".join(conditions_l)})')
        return self

    def set_database(self, database: peewee.Database | peewee.DatabaseProxy) -> Self:
        """
        Set the database to use for this trigger.

        :param database: The database instance
        :return: self for easy chaining
        """
        self._database = database
        return self

    def _get_dialect(self) -> TriggerDialect:
        """
        Get the appropriate dialect based on the database type.

        :return: A dialect instance
        """
        if self._dialect is not None:
            return self._dialect

        if self._database is None:
            # Default to SQLite dialect
            return SQLiteDialect()

        db = self._database
        if isinstance(db, peewee.DatabaseProxy):
            db = db.obj  # Get the actual database from the proxy

        if isinstance(db, peewee.SqliteDatabase):
            self._dialect = SQLiteDialect()
        elif isinstance(db, peewee.MySQLDatabase):
            self._dialect = MySQLDialect()
        elif isinstance(db, peewee.PostgresqlDatabase):
            self._dialect = PostgreSQLDialect()
        else:
            raise UnsupportedDatabaseError(f'Unsupported database type: {type(db)}')

        return self._dialect

    def create(self) -> str:
        """
        Generates the SQL create statement.

        :return: The trigger creation statement.
        :raise MissingSQLStatement: if no SQL statements are provided.
        :raise UnsupportedDatabaseError: if the trigger type is not supported by the database.
        """
        if len(self._sql_list) == 0:
            raise MissingSQLStatement('No SQL statements provided')

        dialect = self._get_dialect()

        # Check if the trigger type is supported
        if not dialect.supports_trigger_type(self.trigger_when, self.trigger_action, self.on_view):
            raise UnsupportedDatabaseError(
                f'Trigger type {self.trigger_when} {self.trigger_action} is not supported by the database'
            )

        # Check if safe create is supported
        if self.safe and not dialect.supports_safe_create():
            # We can either ignore and continue without safe, or raise an error
            # For now, we'll just ignore and continue
            self.safe = False

        # Check if update columns are supported
        if self.update_columns and not dialect.supports_update_of_columns():
            # We can either ignore and continue without column-specific updates, or raise an error
            # For now, we'll ignore and continue
            self.update_columns = []

        # Generate the SQL
        return dialect.create_trigger_sql(self)

    def drop(self, safe: bool = True) -> str:
        """
        Generates the SQL drop statement.

        :param safe: If True, add an IF EXIST. Defaults to True.
        :type safe: bool, Optional
        :return: The drop statement
        :rtype: str
        """
        dialect = self._get_dialect()
        return dialect.drop_trigger_sql(self.trigger_name, safe)



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Module implements a simple timer to measure the execution duration.

Basic usage:

.. code-block:: python

    from mafw import timer

    with Timer() as timer:
        do_long_lasting_operation()

When exiting from the context manager, a message with the duration of the process is printed.

"""

import logging
from datetime import timedelta
from time import perf_counter
from types import TracebackType


def pretty_format_duration(duration_s: float, n_digits: int = 1) -> str:
    """
    Return a formatted version of the duration with increased human readability.

    :param duration_s: The duration to be printed in seconds. If negative, a ValueError exception is raised.
    :type duration_s: float
    :param n_digits: The number of decimal digits to show. Defaults to 1. If negative, a ValueError exception is raised.
    :type n_digits: int, Optional
    :return: The formatted string.
    :rtype: str
    :raises ValueError: if a negative duration or a negative number of digits is provided
    """
    if duration_s < 0:
        raise ValueError(f'Duration ({duration_s}) cannot be a negative value')
    if n_digits < 0:
        raise ValueError('The number of decimal digits cannot be a negative value')
    td = timedelta(seconds=duration_s)
    days = td.days
    hours = td.seconds // 3600
    minutes = (td.seconds - hours * 3600) // 60
    seconds = round((td.seconds - hours * 3600 - minutes * 60) + td.microseconds / 1e6, n_digits)
    time_array = tuple([days, hours, minutes, seconds])
    if time_array <= (0, 0, 0, 0):
        return f'< {time_array[3]:.{n_digits}f} seconds'
    up = ['days', 'hours', 'minutes', 'seconds']
    us = ['day', 'hour', 'minute', 'second']
    msg = ''
    for ti, usi, upi in zip(time_array, us, up):
        if ti == 0:
            pass
        elif ti == 1:
            msg += f'{ti} {usi}, '
        else:
            msg += f'{ti} {upi}, '

    return rreplace(msg.rstrip(', '), ', ', ' and ', 1)


def rreplace(inp_string: str, old_string: str, new_string: str, counts: int) -> str:
    """
    Utility function to replace a substring in a given string a certain number of times starting from the right-most one.

    This function is mimicking the behavior of the string.replace method, but instead of replacing from the left, it
    is replacing from the right.

    :param inp_string: The input string
    :type inp_string: str
    :param old_string: The old substring to be matched. If empty, a ValueError is raised.
    :type old_string: str
    :param new_string: The new substring to be replaced
    :type new_string: str
    :param counts: The number of times the old substring has to be replaced.
    :type counts: int
    :return: The modified string
    :rtype: str
    :raises ValueError: if old_string is empty.
    """
    if old_string == '':
        raise ValueError('old_string cannot be empty')
    li = inp_string.rsplit(old_string, counts)
    return new_string.join(li)


class Timer:
    """
    The timer class.
    """

    def __init__(self, suppress_message: bool = False) -> None:
        """
        Constructor parameter:

        :param suppress_message: A boolean flag to mute the timer
        :type suppress_message: bool
        """
        self._suppress_message = suppress_message
        self.start = 0.0
        self.end = 0.0

    def __enter__(self) -> 'Timer':
        """
        Context manager enter dunder method.

        When an instance of the Timer class is created via the context manager, the start time attribute is set to the
        current time, while the end is set to zero.

        :return: The class instance.
        """
        self.start = perf_counter()
        self.end = 0.0
        return self

    def __exit__(
        self, type_: type[BaseException] | None, value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        """
        Context manager exit dunder method.

        When the method is invoked, the end time is set to the current time, and, if the timer is not muted, a log
        message with the total duration is printed.

        :param type_: Exception type causing the context manager to exit. Defaults to None.
        :type type_: type[BaseException], Optional
        :param value: Exception that caused the context manager to exit. Defaults to None.
        :type value: BaseException, Optional
        :param traceback: Traceback. Defaults to None.
        :type traceback: TracebackType
        """
        self.end = perf_counter()
        if not self._suppress_message:
            logging.getLogger(__name__).info('Total execution time: %s' % self.format_duration())

    @property
    def duration(self) -> float:
        """
        The elapsed time of the timer.

        :return: Elapsed time in seconds.
        """
        return self.end - self.start

    def format_duration(self) -> str:
        """
        Nicely format the timer duration.

        :return: A string with the timer duration in a human-readable formatted string
        :rtype: str
        """
        return pretty_format_duration(self.duration)



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
import pytest

from mafw.enumerators import LoopingStatus, LoopType, ProcessorExitStatus, ProcessorStatus


class TestProcessorExitStatus:
    @pytest.mark.parametrize(
        'status, expected',
        [
            (ProcessorExitStatus.Successful, 'Successful'),
            (ProcessorExitStatus.Failed, 'Failed'),
            (ProcessorExitStatus.Aborted, 'Aborted'),
        ],
    )
    def test_exit_status(self, status, expected):
        assert status.name == expected


class TestProcessorStatus:
    @pytest.mark.parametrize(
        'status, expected',
        [
            (ProcessorStatus.Unknown, 'unknown'),
            (ProcessorStatus.Init, 'initializing'),
            (ProcessorStatus.Start, 'starting'),
            (ProcessorStatus.Run, 'processing'),
            (ProcessorStatus.Finish, 'finishing'),
        ],
    )
    def test_processor_status(self, status, expected):
        assert status.value == expected


class TestLoopingStatus:
    @pytest.mark.parametrize(
        'status, expected',
        [
            (LoopingStatus.Continue, 'Continue'),
            (LoopingStatus.Skip, 'Skip'),
            (LoopingStatus.Abort, 'Abort'),
            (LoopingStatus.Quit, 'Quit'),
        ],
    )
    def test_looping_status(self, status, expected):
        assert status.name == expected


class TestLoopType:
    @pytest.mark.parametrize(
        'loop_type, expected',
        [
            (LoopType.SingleLoop, 'single'),
            (LoopType.ForLoop, 'for_loop'),
            (LoopType.WhileLoop, 'while_loop'),
        ],
    )
    def test_loop_type(self, loop_type, expected):
        assert loop_type.value == expected



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Generic

from hatchling.builders.plugin.interface import BuilderInterface
from hatchling.metadata.core import ProjectMetadata
from hatchling.plugin.manager import PluginManagerBound
from hatchling.plugin.utils import load_plugin_from_script
from hatchling.utils.constants import DEFAULT_BUILD_SCRIPT

if TYPE_CHECKING:
    from hatchling.bridge.app import Application


class CustomBuilder(Generic[PluginManagerBound]):
    PLUGIN_NAME = 'custom'

    def __new__(  # type: ignore[misc]
        cls,
        root: str,
        plugin_manager: PluginManagerBound | None = None,
        config: dict[str, Any] | None = None,
        metadata: ProjectMetadata | None = None,
        app: Application | None = None,
    ) -> BuilderInterface:
        project_metadata = ProjectMetadata(root, plugin_manager, config)

        target_config = project_metadata.hatch.build_targets.get(cls.PLUGIN_NAME, {})
        if not isinstance(target_config, dict):
            message = f'Field `tool.hatch.build.targets.{cls.PLUGIN_NAME}` must be a table'
            raise TypeError(message)

        build_script = target_config.get('path', DEFAULT_BUILD_SCRIPT)
        if not isinstance(build_script, str):
            message = f'Option `path` for builder `{cls.PLUGIN_NAME}` must be a string'
            raise TypeError(message)

        if not build_script:
            message = f'Option `path` for builder `{cls.PLUGIN_NAME}` must not be empty if defined'
            raise ValueError(message)

        path = os.path.normpath(os.path.join(root, build_script))
        if not os.path.isfile(path):
            message = f'Build script does not exist: {build_script}'
            raise OSError(message)

        hook_class = load_plugin_from_script(path, build_script, BuilderInterface, 'builder')  # type: ignore[type-abstract]
        hook = hook_class(root, plugin_manager=plugin_manager, config=config, metadata=metadata, app=app)

        # Always keep the name to avoid confusion
        hook.PLUGIN_NAME = cls.PLUGIN_NAME

        return hook



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Test suite for abstract_plotter module.

This test suite provides comprehensive coverage for the abstract plotter classes and their functionality,
including mocking of external dependencies and testing of various scenarios.
"""

import logging
from unittest.mock import Mock, patch

import peewee
import pytest

from mafw.enumerators import LoopingStatus, LoopType
from mafw.processor_library.abstract_plotter import DataRetriever, FigurePlotter, GenericPlotter, PlotterMeta


class TestPlotterMeta:
    """Test the PlotterMeta metaclass."""

    def test_plotter_meta_inheritance(self):
        """Test that PlotterMeta properly inherits from Protocol and ProcessorMeta."""

        # Test that we can create a class with PlotterMeta
        class TestClass(metaclass=PlotterMeta):
            pass

        assert TestClass.__class__ == PlotterMeta


class TestDataRetriever:
    """Test the DataRetriever abstract base class."""

    def test_data_retriever_is_abstract(self):
        """Test that DataRetriever cannot be instantiated directly."""
        with pytest.raises(TypeError):
            DataRetriever()

    def test_data_retriever_init(self):
        """Test DataRetriever initialization with concrete implementation."""

        class ConcreteDataRetriever(DataRetriever):
            def get_data_frame(self):
                pass

            def patch_data_frame(self):
                pass

            def _attributes_valid(self):
                return True

        # Should be able to instantiate concrete implementation
        retriever = ConcreteDataRetriever()
        assert isinstance(retriever, DataRetriever)

    def test_abstract_methods_must_be_implemented(self):
        """Test that abstract methods must be implemented in subclasses."""

        class IncompleteDataRetriever(DataRetriever):
            pass

        with pytest.raises(TypeError):
            IncompleteDataRetriever()


class TestFigurePlotter:
    """Test the FigurePlotter abstract base class."""

    def test_figure_plotter_is_abstract(self):
        """Test that FigurePlotter cannot be instantiated directly."""
        with pytest.raises(TypeError):
            FigurePlotter()

    def test_figure_plotter_concrete_implementation(self):
        """Test concrete implementation of FigurePlotter."""

        class ConcreteFigurePlotter(FigurePlotter):
            def plot(self):
                pass

            def _attributes_valid(self):
                return True

        plotter = ConcreteFigurePlotter()
        assert isinstance(plotter, FigurePlotter)


class TestGenericPlotter:
    """Test the GenericPlotter class."""

    def test_generic_plotter_metaclass(self):
        """Test that GenericPlotter uses PlotterMeta."""
        assert GenericPlotter.__class__ == PlotterMeta


class TestGenericPlotterIsOutputExisting:
    """Test the is_output_existing method of GenericPlotter."""

    @pytest.fixture
    def plotter_with_db(self):
        """Create a plotter with database connection."""
        plotter = GenericPlotter(looper=LoopType.SingleLoop)
        plotter.name = 'test_plotter'
        plotter._database = Mock()
        return plotter

    @pytest.fixture
    def plotter_without_db(self):
        """Create a plotter without database connection."""
        plotter = GenericPlotter(looper=LoopType.SingleLoop)
        plotter.name = 'test_plotter'
        plotter._database = None
        return plotter

    def test_no_database_connection(self, plotter_without_db, caplog):
        """Test behavior when no database connection is available."""
        with caplog.at_level(logging.WARNING):
            result = plotter_without_db.is_output_existing()

        assert result is False
        assert 'No database connection available' in caplog.text

    @patch('mafw.processor_library.abstract_plotter.PlotterOutput')
    def test_no_existing_output_record(self, mock_plotter_output, plotter_with_db):
        """Test when no output record exists in database."""
        mock_plotter_output.get.side_effect = peewee.DoesNotExist()

        result = plotter_with_db.is_output_existing()

        assert result is False
        mock_plotter_output.get.assert_called_once()

    @patch('mafw.processor_library.abstract_plotter.TriggerDisabler')
    @patch('mafw.processor_library.abstract_plotter.PlotterOutput')
    def test_files_missing(self, mock_plotter_output, trigger_disabler, plotter_with_db):
        """Test when some files are missing."""
        # Mock query result
        mock_query = Mock()
        mock_file1 = Mock()
        mock_file1.exists.return_value = True
        mock_file2 = Mock()
        mock_file2.exists.return_value = False  # This file is missing
        mock_query.filename_list = [mock_file1, mock_file2]
        trigger_disabler.__enter__ = Mock()
        trigger_disabler.__exit__ = Mock()

        mock_plotter_output.get.return_value = mock_query
        mock_plotter_output.delete.return_value.where.return_value.execute.return_value = None

        result = plotter_with_db.is_output_existing()

        assert result is False
        mock_plotter_output.delete.assert_called_once()

    @patch('mafw.processor_library.abstract_plotter.TriggerDisabler')
    @patch('mafw.processor_library.abstract_plotter.PlotterOutput')
    @patch('mafw.processor_library.abstract_plotter.file_checksum')
    def test_checksum_mismatch(self, mock_file_checksum, mock_plotter_output, trigger_disabler, plotter_with_db):
        """Test when checksum doesn't match."""
        # Mock query result
        mock_query = Mock()
        mock_file = Mock()
        mock_file.exists.return_value = True
        mock_query.filename_list = [mock_file]
        mock_query.checksum = 'old_checksum'
        trigger_disabler.__enter__ = Mock()
        trigger_disabler.__exit__ = Mock()

        mock_plotter_output.get.return_value = mock_query
        mock_file_checksum.return_value = 'new_checksum'  # Different checksum
        mock_plotter_output.delete.return_value.where.return_value.execute.return_value = None

        result = plotter_with_db.is_output_existing()

        assert result is False
        mock_plotter_output.delete.assert_called_once()
        mock_file_checksum.assert_called_once_with([mock_file])

    @patch('mafw.processor_library.abstract_plotter.PlotterOutput')
    @patch('mafw.processor_library.abstract_plotter.file_checksum')
    def test_valid_existing_output(self, mock_file_checksum, mock_plotter_output, plotter_with_db):
        """Test when output exists and is valid."""
        # Mock query result
        mock_query = Mock()
        mock_file = Mock()
        mock_file.exists.return_value = True
        mock_query.filename_list = [mock_file]
        mock_query.checksum = 'matching_checksum'

        mock_plotter_output.get.return_value = mock_query
        mock_file_checksum.return_value = 'matching_checksum'  # Same checksum

        result = plotter_with_db.is_output_existing()

        assert result is True
        mock_file_checksum.assert_called_once_with([mock_file])


class TestGenericPlotterProcess:
    """Test the process method of GenericPlotter."""

    @pytest.fixture
    def plotter_with_mocks(self):
        """Create a plotter with all methods mocked."""
        plotter = GenericPlotter(looper=LoopType.SingleLoop)
        plotter.name = 'test_plotter'
        plotter.filter_register = Mock()
        plotter.filter_register.new_only = False

        # Mock all the methods that process() calls
        plotter.is_output_existing = Mock(return_value=False)
        plotter.in_loop_customization = Mock()
        plotter.get_data_frame = Mock()
        plotter.patch_data_frame = Mock()
        plotter.slice_data_frame = Mock()
        plotter.group_and_aggregate_data_frame = Mock()
        plotter.is_data_frame_empty = Mock(return_value=False)
        plotter.plot = Mock()
        plotter.customize_plot = Mock()
        plotter.save = Mock()
        plotter.update_db = Mock()

        return plotter

    def test_process_with_existing_output_and_new_only(self, plotter_with_mocks):
        """Test process method when output exists and new_only is True."""
        plotter_with_mocks.filter_register.new_only = True
        plotter_with_mocks.is_output_existing.return_value = True

        plotter_with_mocks.process()

        # Should return early, not calling other methods
        plotter_with_mocks.is_output_existing.assert_called_once()
        plotter_with_mocks.in_loop_customization.assert_not_called()

    def test_process_full_workflow(self, plotter_with_mocks):
        """Test full process workflow when data frame is not empty."""
        plotter_with_mocks.process()

        # Verify all methods are called in correct order
        plotter_with_mocks.in_loop_customization.assert_called_once()
        plotter_with_mocks.get_data_frame.assert_called_once()
        plotter_with_mocks.patch_data_frame.assert_called_once()
        plotter_with_mocks.slice_data_frame.assert_called_once()
        plotter_with_mocks.group_and_aggregate_data_frame.assert_called_once()
        plotter_with_mocks.is_data_frame_empty.assert_called_once()
        plotter_with_mocks.plot.assert_called_once()
        plotter_with_mocks.customize_plot.assert_called_once()
        plotter_with_mocks.save.assert_called_once()
        plotter_with_mocks.update_db.assert_called_once()

    def test_process_with_empty_data_frame(self, plotter_with_mocks):
        """Test process method when data frame is empty."""
        plotter_with_mocks.is_data_frame_empty.return_value = True

        plotter_with_mocks.process()

        # Should not call plotting methods when data frame is empty
        plotter_with_mocks.is_data_frame_empty.assert_called_once()
        plotter_with_mocks.plot.assert_not_called()
        plotter_with_mocks.customize_plot.assert_not_called()
        plotter_with_mocks.save.assert_not_called()
        plotter_with_mocks.update_db.assert_not_called()


class TestGenericPlotterDefaultMethods:
    """Test default implementations of GenericPlotter methods."""

    @pytest.fixture
    def generic_plotter(self):
        """Create a GenericPlotter instance."""

        plotter = GenericPlotter(looper=LoopType.SingleLoop)
        plotter.name = 'test_plotter'
        plotter.progress_message = ''
        return plotter

    def test_is_data_frame_empty_default(self, generic_plotter):
        """Test default implementation of is_data_frame_empty."""
        result = generic_plotter.is_data_frame_empty()
        assert result is False

    def test_in_loop_customization_default(self, generic_plotter):
        """Test default implementation of in_loop_customization."""
        # Should not raise any exception
        generic_plotter.in_loop_customization()

    def test_get_data_frame_default(self, generic_plotter):
        """Test default implementation of get_data_frame."""
        # Should not raise any exception (it's a pass statement)
        generic_plotter.get_data_frame()

    def test_format_progress_message(self, generic_plotter):
        """Test format_progress_message method."""
        generic_plotter.format_progress_message()
        assert generic_plotter.progress_message == 'test_plotter is working'

    def test_plot_default(self, generic_plotter):
        """Test default implementation of plot."""
        generic_plotter.plot()

    def test_customize_plot_default(self, generic_plotter):
        """Test default implementation of customize_plot."""
        generic_plotter.customize_plot()

    def test_save_default(self, generic_plotter):
        """Test default implementation of save."""
        generic_plotter.save()

    def test_update_db_default(self, generic_plotter):
        """Test default implementation of update_db."""
        generic_plotter.update_db()

    def test_slice_data_frame_default(self, generic_plotter):
        """Test default implementation of slice_data_frame."""
        generic_plotter.slice_data_frame()

    def test_group_and_aggregate_data_frame_default(self, generic_plotter):
        """Test default implementation of group_and_aggregate_data_frame."""
        generic_plotter.group_and_aggregate_data_frame()

    def test_patch_data_frame_default(self, generic_plotter):
        """Test default implementation of patch_data_frame."""
        generic_plotter.patch_data_frame()


class TestGenericPlotterFinish:
    """Test the finish method of GenericPlotter."""

    @pytest.fixture
    def plotter_with_finish_setup(self):
        """Create a plotter set up for testing finish method."""
        plotter = GenericPlotter(looper=LoopType.SingleLoop)
        plotter.looping_status = LoopingStatus.Continue
        plotter._update_plotter_db = Mock()
        return plotter

    @patch('mafw.processor_library.abstract_plotter.Processor.finish')
    def test_finish_with_continue_status(self, mock_super_finish, plotter_with_finish_setup):
        """Test finish method when looping status is Continue."""
        plotter_with_finish_setup.finish()

        plotter_with_finish_setup._update_plotter_db.assert_called_once()
        mock_super_finish.assert_called_once()

    @patch('mafw.processor_library.abstract_plotter.Processor.finish')
    def test_finish_with_other_status(self, mock_super_finish, plotter_with_finish_setup):
        """Test finish method when looping status is not Continue."""
        plotter_with_finish_setup.looping_status = LoopingStatus.Skip

        plotter_with_finish_setup.finish()

        plotter_with_finish_setup._update_plotter_db.assert_not_called()
        mock_super_finish.assert_called_once()


class TestGenericPlotterUpdatePlotterDb:
    """Test the _update_plotter_db method of GenericPlotter."""

    @pytest.fixture
    def plotter_with_db_setup(self):
        """Create a plotter set up for testing database update."""
        plotter = GenericPlotter(looper=LoopType.SingleLoop)
        plotter.name = 'test_plotter'
        plotter.output_filename_list = []
        return plotter

    def test_update_plotter_db_no_database(self, plotter_with_db_setup, caplog):
        """Test _update_plotter_db when no database connection exists."""
        plotter_with_db_setup._database = None

        with caplog.at_level(logging.WARNING):
            plotter_with_db_setup._update_plotter_db()

        assert 'No database connection available' in caplog.text

    def test_update_plotter_db_empty_output_list(self, plotter_with_db_setup):
        """Test _update_plotter_db when output filename list is empty."""
        plotter_with_db_setup._database = Mock()
        plotter_with_db_setup.output_filename_list = []

        # Should return early without doing anything
        plotter_with_db_setup._update_plotter_db()

    @patch('mafw.processor_library.abstract_plotter.PlotterOutput')
    def test_update_plotter_db_success(self, mock_plotter_output, plotter_with_db_setup):
        """Test successful database update."""
        plotter_with_db_setup._database = Mock()
        plotter_with_db_setup.output_filename_list = ['file1.png', 'file2.png']

        # Mock the upsert chain
        mock_upsert = Mock()
        mock_execute = Mock()
        mock_upsert.execute.return_value = mock_execute
        mock_plotter_output.std_upsert.return_value = mock_upsert

        plotter_with_db_setup._update_plotter_db()

        # Verify upsert was called with correct data
        mock_plotter_output.std_upsert.assert_called_once_with(
            {
                'plotter_name': 'test_plotter',
                'filename_list': ['file1.png', 'file2.png'],
                'checksum': ['file1.png', 'file2.png'],
            }
        )
        mock_upsert.execute.assert_called_once()


class TestGenericPlotterParametrized:
    """Parametrized tests for GenericPlotter."""

    @pytest.fixture
    def plotter_base(self):
        """Base plotter for parametrized tests."""

        plotter = GenericPlotter(looper=LoopType.SingleLoop)
        plotter.name = 'test_plotter'
        return plotter

    @pytest.mark.parametrize(
        'looping_status,should_update',
        [
            (LoopingStatus.Continue, True),
            (LoopingStatus.Abort, False),
            (LoopingStatus.Quit, False),
        ],
    )
    @patch('mafw.processor_library.abstract_plotter.Processor.finish')
    def test_finish_with_different_statuses(self, mock_super_finish, plotter_base, looping_status, should_update):
        """Test finish method with different looping statuses."""
        plotter_base.looping_status = looping_status
        plotter_base._update_plotter_db = Mock()

        plotter_base.finish()

        if should_update:
            plotter_base._update_plotter_db.assert_called_once()
        else:
            plotter_base._update_plotter_db.assert_not_called()

        mock_super_finish.assert_called_once()

    @pytest.mark.parametrize(
        'new_only,output_exists,should_continue',
        [
            (True, True, False),  # new_only=True and output exists -> should return early
            (True, False, True),  # new_only=True but no output -> should continue
            (False, True, True),  # new_only=False regardless of output -> should continue
            (False, False, True),  # new_only=False regardless of output -> should continue
        ],
    )
    def test_process_new_only_scenarios(self, plotter_base, new_only, output_exists, should_continue):
        """Test process method with different new_only and output existence scenarios."""
        plotter_base.filter_register = Mock()
        plotter_base.filter_register.new_only = new_only
        plotter_base.is_output_existing = Mock(return_value=output_exists)
        plotter_base.in_loop_customization = Mock()

        plotter_base.process()

        if should_continue:
            plotter_base.in_loop_customization.assert_called_once()
        else:
            plotter_base.in_loop_customization.assert_not_called()


@pytest.mark.integration_test
class TestIntegration:
    """Integration tests for abstract plotter components."""

    def test_mixed_class_creation(self):
        """Test creating a mixed class with DataRetriever and FigurePlotter."""

        class ConcreteDataRetriever(DataRetriever):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.data = ''

            def get_data_frame(self):
                self.data = 'retrieved'

            def patch_data_frame(self):
                self.data = 'patched'

            def _attributes_valid(self):
                return True

        class ConcreteFigurePlotter(FigurePlotter):
            def plot(self):
                self.plotted = True

            def _attributes_valid(self):
                return True

        # remember the generic plotter always at the end!
        class MixedPlotter(ConcreteDataRetriever, ConcreteFigurePlotter, GenericPlotter):
            pass

        mixed = MixedPlotter(looper=LoopType.SingleLoop)
        mixed.get_data_frame()
        mixed.patch_data_frame()
        mixed.plot()

        assert mixed.data == 'patched'
        assert mixed.plotted is True

    def test_abstract_method_enforcement(self):
        """Test that abstract methods are properly enforced."""

        # Test incomplete DataRetriever
        class IncompleteRetriever(DataRetriever):
            def get_data_frame(self):
                pass

            # Missing patch_data_frame and _attributes_valid

        with pytest.raises(TypeError):
            IncompleteRetriever()

        # Test incomplete FigurePlotter
        class IncompletePlotter(FigurePlotter):
            def plot(self):
                pass

            # Missing _attributes_valid

        with pytest.raises(TypeError):
            IncompletePlotter()


# Additional tests for edge cases and error handling
class TestEdgeCases:
    """Test edge cases and error handling."""

    @pytest.fixture
    def plotter_edge_case(self):
        """Create plotter for edge case testing."""

        plotter = GenericPlotter(looper=LoopType.SingleLoop)
        plotter.name = 'edge_case_plotter'
        return plotter

    def test_logging_configuration(self, plotter_edge_case):
        """Test logging is properly configured."""
        import mafw.processor_library.abstract_plotter as module

        assert hasattr(module, 'log')
        assert isinstance(module.log, logging.Logger)

    @patch('mafw.processor_library.abstract_plotter.PlotterOutput')
    def test_database_exception_handling(self, mock_plotter_output, plotter_edge_case):
        """Test handling of unexpected database exceptions."""
        plotter_edge_case._database = Mock()

        # Simulate unexpected database error
        mock_plotter_output.get.side_effect = Exception('Unexpected DB error')

        # Should not raise exception, but return False
        with pytest.raises(Exception):
            plotter_edge_case.is_output_existing()



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Module implements the basic Processor class, the ProcessorList and all helper classes to achieve the core
functionality of the MAFw.
"""

from __future__ import annotations

import contextlib
import inspect
import logging
import warnings
from collections.abc import Callable, Iterator
from copy import copy
from functools import wraps
from itertools import count
from typing import Any, Collection, Generic, Iterable, Self, SupportsIndex, TypeVar, Union, cast

import peewee
from peewee import Database

# noinspection PyUnresolvedReferences
from playhouse.db_url import connect

import mafw.db.db_filter
from mafw.active import Active
from mafw.db.db_model import database_proxy
from mafw.db.db_types import PeeweeModelWithMeta
from mafw.db.std_tables import standard_tables
from mafw.enumerators import LoopingStatus, LoopType, ProcessorExitStatus, ProcessorStatus
from mafw.mafw_errors import (
    AbortProcessorException,
    MissingDatabase,
    MissingOverloadedMethod,
    MissingSuperCall,
    ProcessorParameterError,
)
from mafw.timer import Timer, pretty_format_duration
from mafw.tools.regexp import extract_protocol
from mafw.ui.abstract_user_interface import UserInterfaceBase
from mafw.ui.console_user_interface import ConsoleInterface

log = logging.getLogger(__name__)

ParameterType = TypeVar('ParameterType')
"""Generic variable type for the :class:`ActiveParameter` and :class:`PassiveParameter`."""


def validate_database_conf(database_conf: dict[str, Any] | None = None) -> dict[str, Any] | None:
    """
    Validates the database configuration.

    :param database_conf: The input database configuration. Defaults to None.
    :type database_conf: dict, Optional
    :return: Either the validated database configuration or None if it is invalid.
    :rtype: dict, None
    """
    if database_conf is None:
        return None

    # dict is mutable, if I change it inside the function, I change it also outside.
    conf = database_conf.copy()

    if 'DBConfiguration' in conf:
        # the database_conf is the steering file. Extract the DBConfiguration table
        conf = conf['DBConfiguration']
    required_fields = ['URL']
    if all([field in conf for field in required_fields]):
        return database_conf
    else:
        return None


class PassiveParameter(Generic[ParameterType]):
    """
    A processor parameter that can be registered and configured.

    For Processors to perform their analytical task, it may be necessary to have some configurable parameters,
    like a DB input table or the output folder or a numeric parameter.

    The name of the parameter must be unique within the Processor scope and a valid python identifier.

    When defined as a PassiveParameter, an instance variable can have a default value if the user did not provide
    one, and it is much easier to configure via a configuration file.

    A parameter for which only a default value is provided is automatically considered *optional*.

    If both the value and the default value are not provided, an exception is raised.

    This class is working behind the scene, that is why it is named **passive**. The user will very likely add class
    instances of :class:`ActiveParameter`, that are publicly exposed in the processor class namespace,
    and an PassiveParameter will automatically added to the class. To access this passive parameter the user
    can use the :meth:`Processor.get_parameter` using the name as key. The :attr:`value` of the passive parameter is
    always accessible using calling the corresponding ActiveParameter.

    .. seealso::

        An explanation on how processor parameters work and should be used is given in :ref:`Understanding processor
        parameters <parameters>`
    """

    def __init__(
        self, name: str, value: ParameterType | None = None, default: ParameterType | None = None, help_doc: str = ''
    ):
        """
        Constructor parameters:

        :param name: The name of the parameter. It must be a valid python identifier.
        :type name: str
        :param value: The set value of the parameter. If None, then the default value will be used. Defaults to None.
        :type value: ParameterType, Optional
        :param default: The default value for the parameter. It is used if the :attr:`value` is not provided. Defaults to None.
        :type default: ParameterType, Optional
        :param help_doc: A brief explanation of the parameter.
        :type help_doc: str, Optional
        :raises ProcessorParameterError: if both `value` and `default` are not provided or if `name` is not a valid identifier.
        """
        if not name.isidentifier():
            raise ProcessorParameterError(f'{name} is not a valid python identifier.')

        self.name = name

        if value is not None:
            self._value: ParameterType = value
            self._is_set = True
            self._is_optional = False
        elif default is not None:
            self._value = default
            self._is_set = False
            self._is_optional = True
        else:
            raise ProcessorParameterError('Processor parameter cannot have both value and default value set to None')

        self.doc = help_doc

    def __rich_repr__(self) -> Iterator[Any]:
        yield 'name', self.name
        yield 'value', self.value, None
        yield 'help_doc', self.doc, ''

    @property
    def is_set(self) -> bool:
        """
        Property to check if the value has been set.

        It is useful for optional parameter to see if the current value is the default one, or if the user set it.
        """
        return self._is_set

    @property
    def value(self) -> ParameterType:
        """
        Gets the parameter value.

        :return: The parameter value.
        :rtype: ParameterType
        :raises ProcessorParameterError: if both value and default were not defined.
        """
        return self._value

    @value.setter
    def value(self, value: ParameterType) -> None:
        """
        Sets the parameter value.

        :param value: The value to be set.
        :type value: ParameterType
        """
        self._value = value
        self._is_set = True

    @property
    def is_optional(self) -> bool:
        """
        Property to check if the parameter is optional.

        :return: True if the parameter is optional
        :rtype: bool
        """
        return self._is_optional

    def __repr__(self) -> str:
        args = ['name', 'value', 'doc']
        values = [getattr(self, arg) for arg in args]
        return '{klass}({attrs})'.format(
            klass=self.__class__.__name__,
            attrs=', '.join('{}={!r}'.format(k, v) for k, v in zip(args, values)),
        )


F = TypeVar('F', bound=Callable[..., Any])
"""Type variable for generic callable with any return value."""


def ensure_parameter_registration(func: F) -> F:
    """Decorator to ensure that before calling `func` the processor parameters have been registered."""

    @wraps(func)
    def wrapper(*args: Processor, **kwargs: Any) -> F:
        # the first positional arguments must be self
        if len(args) == 0:
            raise ProcessorParameterError(
                'Attempt to apply the ensure_parameter_registration to something different to a Processor subclass.'
            )
        self = args[0]
        if not isinstance(self, Processor):
            raise ProcessorParameterError(
                'Attempt to apply the ensure_parameter_registration to something different to a Processor subclass.'
            )
        if self._parameter_registered is False:
            self._register_parameters()
        return cast(F, func(*args, **kwargs))

    return cast(F, wrapper)


class ActiveParameter(Generic[ParameterType]):
    r"""
    The public interface to the processor parameter.

    The behaviour of a :class:`Processor` can be customized by using processor parameters. The value of these
    parameters can be either set via a configuration file or directly when creating the class.

    If the user wants to benefit from this facility, they have to add in the instance of the Processor subclass an
    ActiveParameter instance in this way:

    .. code-block::

        class MyProcessor(Processor):

            # this is the input folder
            input_folder = ActiveParameter('input_folder', Path(r'C:\'), help_doc='This is where to look for input files')

            def __init__(self, *args, **kwargs):
                super().__init(*args, **kwargs)

                # change the input folder to something else
                self.input_folder = Path(r'D:\data')

                # get the value of the parameter
                print(self.input_folder)

    The ActiveParameter is a `descriptor <https://docs.python.org/3/glossary.html#term-descriptor>`_, it means that
    when you create one of them, a lot of work is done behind the scene.

    In simple words, a processor parameter is made by two objects: a public interface where the user can easily
    access the value of the parameter and a private interface where all other information (default, documentation...)
    is also stored.

    The user does not have to take care of all of this. When a new ActiveParameter instance is added to the class as
    in the code snipped above, the private interface is automatically created and will stay in the class instance
    until the end of the class lifetime.

    To access the private interface, the user can use the :meth:`Processor.get_parameter` method using the parameter
    name as a key.

    .. seealso::

        The private counter part in the :class:`PassiveParameter`.

    """

    def __init__(
        self, name: str, value: ParameterType | None = None, default: ParameterType | None = None, help_doc: str = ''
    ):
        """
        Constructor parameters:

        :param name: The name of the parameter.
        :type name: str
        :param value: The initial value of the parameter. Defaults to None.
        :type value: ParameterType, Optional
        :param default: The default value of the parameter, to be used when ``value`` is not set., Defaults to None.
        :type default: ParameterType, Optional
        :param help_doc: An explanatory text describing the parameter.
        :type help_doc: str, Optional
        """

        self._value = value
        self._default = default
        self._help_doc = help_doc
        self._external_name = name

    def __set_name__(self, owner: type[Processor], name: str) -> None:
        self.public_name = name
        self.private_name = f'param_{name}'
        self._owner = owner

        if not hasattr(owner, self.private_name):
            setattr(
                owner,
                self.private_name,
                PassiveParameter(self._external_name, self._value, self._default, self._help_doc),
            )

    def __get__(self, obj: Processor, obj_type: type[Processor]) -> ActiveParameter[ParameterType] | ParameterType:
        if obj is None:
            return self

        param: PassiveParameter[ParameterType] = getattr(obj, self.private_name)
        return param.value

    def __set__(self, obj: Processor, value: ParameterType) -> None:
        param: PassiveParameter[ParameterType] = getattr(obj, self.private_name)
        param.value = value


class ProcessorMeta(type):
    """A metaclass to implement the post-init method."""

    def __call__(cls, *args: Any, **kwargs: Any) -> 'ProcessorMeta':
        obj = type.__call__(cls, *args, **kwargs)
        obj.__post_init__()
        return cast(ProcessorMeta, obj)


# noinspection PyProtectedMember
class Processor(metaclass=ProcessorMeta):
    """
    The basic processor.

    A very comprehensive description of what a Processor does and how it works is available at :ref:`doc_processor`.
    """

    processor_status = Active(ProcessorStatus.Unknown)
    """Processor execution status"""

    looping_status = Active(LoopingStatus.Continue)
    """Looping modifier"""

    progress_message: str = f'{__qualname__} is working'
    """Message displayed to show the progress. 
    
    It can be customized with information about the current item in the loop by overloading the 
    :meth:`format_progress_message`."""

    _ids = count(0)
    """A counter for all processor instances"""

    def __init__(
        self,
        name: str | None = None,
        description: str | None = None,
        config: dict[str, Any] | None = None,
        looper: LoopType | str = LoopType.ForLoop,
        user_interface: UserInterfaceBase | None = None,
        timer: Timer | None = None,
        timer_params: dict[str, Any] | None = None,
        database: Database | None = None,
        database_conf: dict[str, Any] | None = None,
        remove_orphan_files: bool = True,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Constructor parameters

        :param name: The name of the processor. If None is provided, the class name is used instead. Defaults to None.
        :type name: str, Optional
        :param description: A short description of the processor task. Defaults to the processor name.
        :type description: str, Optional
        :param config: A configuration dictionary for this processor. Defaults to None.
        :type config: dict, Optional
        :param looper: Enumerator to define the looping type. Defaults to LoopType.ForLoop
        :type looper: LoopType, Optional
        :param user_interface: A user interface instance to be used by the processor to interact with the user.
        :type user_interface: UserInterfaceBase, Optional
        :param timer: A timer object to measure process duration.
        :type timer: Timer, Optional
        :param timer_params: Parameters for the timer object.
        :type timer_params: dict, Optional
        :param database: A database instance. Defaults to None.
        :type database: Database, Optional
        :param database_conf: Configuration for the database. Default to None.
        :type database_conf: dict, Optional
        :param remove_orphan_files: Boolean flag to remove files on disc without a reference to the database.
            See :ref:`std_tables` and :meth:`~mafw.processor.Processor._remove_orphan_files`. Defaults to True
        :type remove_orphan_files: bool, Optional
        :param kwargs: Keyword arguments that can be used to set processor parameters.
        """

        self.name = name or self.__class__.__name__
        """The name of the processor."""

        self.unique_id = next(self._ids)
        """A unique identifier representing how many instances of Processor has been created."""

        self.description = description or self.name
        """A short description of the processor task."""

        self.item: Any = None
        """The current item of the loop."""

        self.processor_exit_status = ProcessorExitStatus.Successful
        """Processor exit status"""

        self.loop_type: LoopType = LoopType(looper)
        """
        The loop type. 
        
        The value of this parameter can also be changed by the :func:`~mafw.decorators.execution_workflow` decorator 
        factory.  
        
        See :class:`~mafw.enumerators.LoopType` for more details.
        """

        # private attributes
        self._config = config or {}
        self._processor_parameters: dict[str, PassiveParameter[ParameterType]] = {}  # type: ignore
        # wait for answer from SO
        self._parameter_registered = False
        self._kwargs = kwargs

        # loops attributes
        self._i_item: int = -1
        self._n_item: int | None = -1
        self._process_durations: list[float] = []

        # resource stack
        self._resource_stack: contextlib.ExitStack
        self._resource_acquisition: bool = True

        # processor timer
        self.timer: Timer | None = timer
        self._timer_parameters: dict[str, Any] = timer_params or {}

        # user interface
        if user_interface is None:
            self._user_interface: UserInterfaceBase = ConsoleInterface()
        else:
            self._user_interface = user_interface

        # database stuff
        self._database: peewee.Database | None = database
        self._database_conf: dict[str, Any] | None = validate_database_conf(database_conf)
        self.filter_register: mafw.db.db_filter.FilterRegister = mafw.db.db_filter.FilterRegister()
        """The DB filter register of the Processor."""
        self.remove_orphan_files: bool = remove_orphan_files
        """The flag to remove or protect the orphan files. Defaults to True"""

        # sub-classing stuff
        # todo:
        #   should we make it a class attribute instead of an instance attribute?
        self._methods_to_be_checked_for_super = [('start', Processor), ('finish', Processor)]
        """
        List of methods to be checked for super inclusion.
        
        It is a list of tuple, with the first element the name of the method to be checked and the second the base 
        class to the be compared.
        """

    def __post_init__(self) -> None:
        self._register_parameters()
        self._load_parameter_configuration()
        self._overrule_kws_parameters()
        self._check_method_overload()
        self._check_method_super()
        self.processor_status = ProcessorStatus.Init

    def _register_parameters(self) -> None:
        if self._parameter_registered:
            return

        for attr_name in dir(self):
            if attr_name in [
                'database'
            ]:  # database is a property that might be not yet initialised and will throw an exception
                continue
            attr = getattr(self, attr_name)
            if isinstance(attr, PassiveParameter):
                if attr.name in self._processor_parameters:
                    raise ProcessorParameterError(f'Duplicated parameter name ({attr_name}.')
                self._processor_parameters[attr.name] = attr
        self._parameter_registered = True

    def _reset_parameters(self) -> None:
        self._processor_parameters = {}
        self._parameter_registered = False
        self._register_parameters()

    @ensure_parameter_registration
    def _load_parameter_configuration(self) -> None:
        original_config = copy(self._config)
        flt_list = []
        if 'GlobalFilter' in original_config:
            self.filter_register._global_filter = original_config['GlobalFilter']

        # we need to check if the configuration object is of type 1 or type 2
        # this is option 1
        if self.name in self._config:
            self._config = self._config[self.name]
        # otherwise, we assume is option 2

        for key, value in self._config.items():
            if key in self._processor_parameters:
                type_: ParameterType = type(self.get_parameter(key).value)  # type: ignore # wait for answer from SO
                self.set_parameter_value(key, type_(value))  # type: ignore # no idea how to fix it, may be linked with above
            elif key == 'Filter':
                # we got a filter table!
                # it should contain one table for each model
                # we add all the names to a list for deferred initialisation
                flt_table = self._config[key]
                flt_list.extend([f'{self.name}.Filter.{model}' for model in flt_table])

        # only now, after the configuration file has been totally read, we can do the real filter initialisation.
        # This is to be sure that if there were a GlobalFilter table, this has been read.
        # The global filter region will be used as a starting point for the construction of a new filter (default
        # parameter in the from_conf class method).
        for flt_name in flt_list:
            model_name = flt_name.split('.')[-1]
            self.filter_register[model_name] = mafw.db.db_filter.Filter.from_conf(
                flt_name, original_config, original_config.get('GlobalFilter', None)
            )

    @ensure_parameter_registration
    def _overrule_kws_parameters(self) -> None:
        for key, value in self._kwargs.items():
            if key in self._processor_parameters:
                type_: ParameterType = type(self.get_parameter(key).value)  # type: ignore # wait for answer from SO
                self.set_parameter_value(key, type_(value))  # type: ignore # no idea how to fix it, may be linked with above

    def _check_method_overload(self) -> None:
        """
        Check if the user overloaded the required methods.

        Depending on the loop type, the user must overload different methods.
        This method is doing the check and if the required methods are not overloaded a warning is emitted.
        """
        methods_dict: dict[LoopType, list[str]] = {
            LoopType.WhileLoop: ['while_condition'],
            LoopType.ForLoop: ['get_items'],
        }
        required_methods: list[str] = methods_dict.get(self.loop_type, [])
        for method in required_methods:
            if getattr(type(self), method) == getattr(Processor, method):
                warnings.warn(
                    MissingOverloadedMethod(
                        '%s was not overloaded. The process execution workflow might not work.' % method
                    )
                )

    def _check_method_super(self) -> None:
        """
        Check if some specific methods are calling their super.

        For some specific methods (for example: start and finish), the user should always call their super method.
        This method verifies that the user implementation of these methods is including a super call, otherwise a
        warning is emitted to inform the user about the problem and possible misbehaviour of the processor.

        The list of methods to be verified is stored in a private class attribute
        :attr:`~._methods_to_be_checked_for_super` as a list of tuples, made by the name of the methods to be
        verified and the base class for comparison. The base class is required because Processor subclasses may be
        extending this list with methods that are not present in the base Processor. See, for example, the
        :meth:`~.GenericPlotter.patch_data_frame` that is required to have a super call, but it is not present in the
        base Processor.

        """
        for method, base in self._methods_to_be_checked_for_super:
            # first check if the user overloaded the method.
            if getattr(type(self), method) != getattr(base, method):
                # if the method is the start method, then it might be that the user decorated the class with the @database_required decorator,
                # so it looks different, but it is not
                if method == 'start':
                    sub_start_src = inspect.getsource(getattr(type(self), method))
                    base_start_src = inspect.getsource(getattr(base, method))
                    if sub_start_src == base_start_src:
                        # it is actually the same method even though the function object is different, then
                        # there is no need to check for the super_call
                        continue

                # let's check if in the overloaded method there is super calls
                super_call = f'super().{method}'
                method_object = getattr(type(self), method)

                # this is the overloaded method source code.
                method_source_code = inspect.getsource(method_object)
                # we split the whole code in lines
                code_lines = method_source_code.split('\n')
                # we remove all comments, because the user may have commented out the super
                code_lines = [line.strip() for line in code_lines if not line.strip().startswith('#')]
                # we rebuild the whole source code, without indentation and comments.
                method_source_code = '\n'.join(code_lines)

                # check if the super call is in the source. if not then emit a warning
                if super_call not in method_source_code:
                    warnings.warn(
                        MissingSuperCall(
                            'The overloaded %s is not invoking its super method. The processor might not work.' % method
                        )
                    )

    @ensure_parameter_registration
    def dump_parameter_configuration(self, option: int = 1) -> dict[str, Any]:
        """
        Dumps the processor parameter values in a dictionary.

        The snipped below explains the meaning of `option`.

        .. code-block:: python

            # option 1
            conf_dict1 = {
                'Processor': {'param1': 5, 'input_table': 'my_table'}
            }

            # option 2
            conf_dict2 = {'param1': 5, 'input_table': 'my_table'}

        :param option: Select the dictionary style. Defaults to 1.
        :type option: int, Optional
        :return: A parameter configuration dictionary.
        :rtype: dict
        """
        inner_dict = {}
        for key, value in self._processor_parameters.items():
            inner_dict[key] = value.value

        if option == 1:
            outer_dict = {self.name: inner_dict}
        elif option == 2:
            outer_dict = inner_dict
        else:
            log.warning('Unknown option %s. Using option 2' % option)
            outer_dict = inner_dict
        return outer_dict

    @ensure_parameter_registration
    def get_parameter(self, name: str) -> PassiveParameter[ParameterType]:
        """
        Gets the processor parameter named name.

        :param name: The name of the parameter.
        :type name: str
        :return: The processor parameter
        :rtype: PassiveParameter
        :raises ProcessorParameterError: If a parameter with `name` is not registered.
        """
        if name in self._processor_parameters:
            return self._processor_parameters[name]
        raise ProcessorParameterError(f'No parameter ({name}) found for {self.name}')

    @ensure_parameter_registration
    def get_parameters(self) -> dict[str, PassiveParameter[ParameterType]]:
        """
        Returns the full dictionary of registered parameters for this processor.

        Useful when dumping the parameter specification in a configuration file, for example.

        :return: The dictionary with the registered parameters.
        :rtype: dict[str, PassiveParameter[ParameterType]
        """
        return self._processor_parameters

    @ensure_parameter_registration
    def delete_parameter(self, name: str) -> None:
        """
        Deletes a processor parameter.

        :param name: The name of the parameter to be deleted.
        :type name: str
        :raises ProcessorParameterError: If a parameter with `name` is not registered.
        """
        if name in self._processor_parameters:
            del self._processor_parameters[name]
        else:
            raise ProcessorParameterError(f'No parameter ({name}) found for {self.name}')

    @ensure_parameter_registration
    def set_parameter_value(self, name: str, value: ParameterType) -> None:
        """
        Sets the value of a processor parameter.

        :param name: The name of the parameter to be deleted.
        :type name: str
        :param value: The value to be assigned to the parameter.
        :type value: ParameterType
        :raises ProcessorParameterError: If a parameter with `name` is not registered.
        """
        if name in self._processor_parameters:
            self._processor_parameters[name].value = value
        else:
            raise ProcessorParameterError(f'No parameter ({name}) found for {self.name}')

    def get_filter(self, model_name: str) -> mafw.db.db_filter.Filter:
        """
        Returns a registered :class:`~mafw.db.db_filter.Filter` via the model name.

        If a filter for the provided model_name does not exist, a KeyError is raised.

        :param model_name: The model name for which the filter will be returned.
        :type model_name: str
        :return: The registered filter
        :rtype: mafw.db.db_filter.Filter
        :raises: KeyError is a filter with the give name is not found.
        """
        return self.filter_register[model_name]

    def on_processor_status_change(self, old_status: ProcessorStatus, new_status: ProcessorStatus) -> None:
        """
        Callback invoked when the processor status is changed.

        :param old_status: The old processor status.
        :type old_status: ProcessorStatus
        :param new_status: The new processor status.
        :type new_status: ProcessorStatus
        """
        self._user_interface.change_of_processor_status(self.name, old_status, new_status)

    def on_looping_status_set(self, status: LoopingStatus) -> None:
        """
        Call back invoked when the looping status is set.

        The user can overload this method according to the needs.

        :param status: The set looping status.
        :type status: LoopingStatus
        """
        if status == LoopingStatus.Skip:
            log.warning('Skipping item %s' % self.i_item)
        elif status == LoopingStatus.Abort:
            log.error('Looping has been aborted')
        elif status == LoopingStatus.Quit:
            log.warning('Looping has been quit')

    def format_progress_message(self) -> None:
        """Customizes the progress message with information about the current item.

        The user can overload this method in order to modify the message being displayed during the process loop with
        information about the current item.

        The user can access the current value, its position in the looping cycle and the total number of items using
        :attr:`.Processor.item`, :obj:`.Processor.i_item` and :obj:`.Processor.n_item`.
        """
        pass

    @property
    def i_item(self) -> int:
        """The enumeration of the current item being processed."""
        return self._i_item

    @i_item.setter
    def i_item(self, value: int) -> None:
        self._i_item = value

    @property
    def n_item(self) -> int | None:
        """The total number of items to be processed or None for an undefined loop"""
        return self._n_item

    @n_item.setter
    def n_item(self, value: int | None) -> None:
        self._n_item = value

    @property
    def unique_name(self) -> str:
        """Returns the unique name for the processor."""
        return f'{self.name}_{self.unique_id}'

    @property
    def local_resource_acquisition(self) -> bool:
        """
        Checks if resources should be acquired locally.

        When the processor is executed in stand-alone mode, it is responsible to acquire and release its own external
        resources, but when it is executed from a ProcessorList, then is a good practice to share and distribute
        resources among the whole processor list. In this case, resources should not be acquired locally by the
        single processor, but from the parent execution context.

        :return: True if resources are to be acquired locally by the processor. False, otherwise.
        :rtype: bool
        """
        return self._resource_acquisition

    @local_resource_acquisition.setter
    def local_resource_acquisition(self, flag: bool) -> None:
        self._resource_acquisition = flag

    @property
    def database(self) -> peewee.Database:
        """
        Returns the database instance

        :return: A database object.
        :raises MissingDatabase: If the database connection has not been established.
        """
        if self._database is None:
            raise MissingDatabase('Database connection not initialized')
        return self._database

    def execute(self) -> None:
        """Execute the processor tasks.

        This method works as a dispatcher, reassigning the call to a more specific execution implementation depending
        on the :attr:`~mafw.processor.Processor.loop_type`.
        """
        dispatcher: dict[LoopType, Callable[[], None]] = {
            LoopType.SingleLoop: self._execute_single,
            LoopType.ForLoop: self._execute_for_loop,
            LoopType.WhileLoop: self._execute_while_loop,
        }
        dispatcher[self.loop_type]()

    def _execute_single(self) -> None:
        """Execute the processor in single mode.

        **Private method**. Do not overload nor invoke it directly. The :meth:`execute` method will call the
        appropriate implementation depending on the processor LoopType.
        """
        with contextlib.ExitStack() as self._resource_stack:
            self.acquire_resources()
            self.start()
            self.processor_status = ProcessorStatus.Run
            self.process()
            self.finish()

    def _execute_for_loop(self) -> None:
        """Executes the processor within a for loop.

        **Private method**. Do not overload nor invoke it directly. The :meth:`execute` method will call the
        appropriate implementation depending on the processor LoopType.
        """

        with contextlib.ExitStack() as self._resource_stack:
            self.acquire_resources()
            self.start()

            # get the input item list and filter it
            item_list = self.get_items()

            # get the total number of items.
            self.n_item = len(item_list)

            # turn the processor status to run
            self.processor_status = ProcessorStatus.Run

            # create a new task in the progress bar interface
            self._user_interface.create_task(self.unique_name, self.description, completed=0, total=self.n_item)

            # start the looping
            for self.i_item, self.item in enumerate(item_list):
                # set the looping status to Continue. The user may want to change it in the process.
                self.looping_status = LoopingStatus.Continue

                # send a message to the user interface
                self.format_progress_message()
                self._user_interface.display_progress_message(self.progress_message, self.i_item, self.n_item, 0.1)

                # wrap the execution in a timer to measure how long it took for statistical reasons.
                with Timer(suppress_message=True) as timer:
                    self.process()
                self._process_durations.append(timer.duration)

                # modify the loop depending on the looping status
                if self.looping_status == LoopingStatus.Continue:
                    self.accept_item()
                elif self.looping_status == LoopingStatus.Skip:
                    self.skip_item()
                else:  # equiv to if self.looping_status in [LoopingStatus.Abort, LoopingStatus.Quit]:
                    break

                # update the progress bar
                self._user_interface.update_task(self.unique_name, self.i_item + 1, 1, self.n_item)
            self._user_interface.update_task(self.unique_name, completed=self.n_item, total=self.n_item)

            self.finish()

    def _execute_while_loop(self) -> None:
        """Executes the processor within a while loop.

        **Private method**. Do not overload nor invoke it directly. The :meth:`execute` method will call the
        appropriate implementation depending on the processor LoopType.
        """
        # it is a while loop, so a priori we don't know how many iterations we will have, nevertheless, we
        # can have a progress bar with 'total' set to None, so that it goes in the so-called indeterminate
        # progress. See https://rich.readthedocs.io/en/stable/progress.html#indeterminate-progress
        # we initialise n_item outside the loop, because it is possible that the user has a way to define n_item
        # and he can do it within the loop.
        self.n_item = None
        with contextlib.ExitStack() as self._resource_stack:
            self.acquire_resources()
            self.start()

            # turn the processor status to run
            self.processor_status = ProcessorStatus.Run

            self._user_interface.create_task(self.unique_name, self.description, completed=0, total=self.n_item)

            # we are ready to start the looping. For statistics, we can count the iterations.
            self.i_item = 0
            while self.while_condition():
                # set the looping status to Continue. The user may want to change it in the process method.
                self.looping_status = LoopingStatus.Continue

                # send a message to the user interface
                self.format_progress_message()
                self._user_interface.display_progress_message(
                    self.progress_message, self.i_item, self.n_item, frequency=0.1
                )

                # wrap the execution in a timer to measure how long it too for statistical reasons.
                with Timer(suppress_message=True) as timer:
                    self.process()
                self._process_durations.append(timer.duration)

                # modify the loop depending on the looping status
                if self.looping_status == LoopingStatus.Continue:
                    self.accept_item()
                elif self.looping_status == LoopingStatus.Skip:
                    self.skip_item()
                else:  # equiv to if self.looping_status in [LoopingStatus.Abort, LoopingStatus.Quit]:
                    break

                # update the progress bar. if self.n_item is still None, then the progress bar will show indeterminate
                # progress.
                self._user_interface.update_task(self.unique_name, self.i_item + 1, 1, self.n_item)

            # now that the loop is finished, we know how many elements we processed
            if self.n_item is None:
                self.n_item = self.i_item
            self._user_interface.update_task(self.unique_name, completed=self.n_item, total=self.n_item)

            self.finish()

    def acquire_resources(self) -> None:
        """
        Acquires resources and add them to the resource stack.

        The whole body of the :meth:`execute` method is within a context structure. The idea is that if any part of
        the code inside should throw an exception that breaking the execution, we want to be sure that all stateful
        resources are properly closed.

        Since the number of resources may vary, the variable number of nested `with` statements has been replaced by
        an `ExitStack <https://docs.python.org/3/library/contextlib.html#contextlib.ExitStack>`_. Resources,
        like open files, timers, db connections, need to be added to the resource stacks in this method.

        In the case a processor is being executed within a :class:`~mafw.processor.ProcessorList`, then some resources might be shared, and
        for this reason they are not added to the stack. This selection can be done via the private
        :attr:`local_resource_acquisition`. This is normally True, meaning that the processor will handle its resources
        independently, but when the processor is executed from a :class:`~mafw.processor.ProcessorList`, this flag is automatically turned to
        False.

        If the user wants to add additional resources, he has to overload this method calling the super to preserve
        the original resources. If he wants to have shared resources among different processors executed from inside
        a processor list, he has to overload the :class:`~mafw.processor.ProcessorList` class as well.
        """
        # Both the timer and the user interface will be added to the processor resource stack only if the processor is
        # set to acquire its own resources.
        # The timer and the user interface have in-built enter and exit method.
        if self._resource_acquisition:
            self.timer = self._resource_stack.enter_context(Timer(**self._timer_parameters))
            self._resource_stack.enter_context(self._user_interface)

        # For the database it is is a bit different.
        if self._database is None and self._database_conf is None:
            # no database, nor  configuration.
            # we cannot do anything
            pass
        elif self._database is None and self._database_conf is not None:
            # no db, but we got a configuration.
            # we can make a db.
            # This processor will try to make a valid connection, and in case it succeeds, it will add the database to
            # the resource stack.
            # The database has an enter method, but it is to generate transaction.
            # We will add the database.close via the callback method.
            if 'DBConfiguration' in self._database_conf:
                conf = self._database_conf['DBConfiguration']  # type1
            else:
                conf = self._database_conf  # type2

            # guess the database type from the URL
            protocol = extract_protocol(conf.get('URL'))

            # build the connection parameter
            # in case of sqlite, we add the pragmas group as well
            connection_parameters = {}
            if protocol == 'sqlite':
                connection_parameters['pragmas'] = conf.get('pragmas', {})
            for key, value in conf.items():
                if key not in ['URL', 'pragmas']:
                    connection_parameters[key] = value

            self._database = connect(conf.get('URL'), **connection_parameters)  # type: ignore # peewee is not returning a DB
            self._resource_stack.callback(self._database.close)
            try:
                self._database.connect()
            except peewee.OperationalError as e:
                log.critical('Unable to connect to %s', self._database_conf.get('URL'))
                raise e
            database_proxy.initialize(self._database)
            self.database.create_tables(standard_tables.values())
            for table in standard_tables.values():
                table.init()

        else:  # equivalent to: if self._database is not None:
            # we got a database, so very likely we are inside a processor list
            # the connection has been already set and the initialisation as well.
            # nothing else to do here.
            # do not put the database in the exit stack. who create it has also to close it.
            pass

    def start(self) -> None:
        """
        Start method.

        The user can overload this method, including all steps that should be performed at the beginning of the
        operation.

        If the user decides to overload it, it should include a call to the super method.
        """
        self.processor_status = ProcessorStatus.Start
        self._remove_orphan_files()

    def get_items(self) -> Collection[Any]:
        """
        Returns the item collections for the processor loop.

        This method must be overloaded for the processor to work. Generally, this is getting a list of rows from the
        database, or a list of files from the disk to be processed.

        :return: A collection of items for the loop
        :rtype: Collection[Any]
        """
        return []

    def while_condition(self) -> bool:
        """
        Return the while condition

        :return: True if the while loop has to continue, false otherwise.
        :rtype: bool
        """
        return False

    def process(self) -> None:
        """
        Processes the current item.

        This is the core of the Processor, where the user has to define the calculations required.
        """
        pass

    def accept_item(self) -> None:
        """
        Does post process actions on a successfully processed item.

        Within the :meth:`process`, the user left the looping status to Continue, so it means that everything looks
        good and this is the right place to perform database updates or file savings.

        .. seealso:
            Have a look at :meth:`skip_item` for what to do in case something went wrong.
        """
        pass

    def skip_item(self) -> None:
        """
        Does post process actions on a *NOT* successfully processed item.

        Within the :meth:`process`, the user set the looping status to Skip, so it means that something went wrong
        and here corrective actions can be taken if needed.

        .. seealso:
            Have a look at :meth:`accept_item` for what to do in case everything was OK.
        """
        pass

    def finish(self) -> None:
        """
        Concludes the execution.

        The user can reimplement this method if there are some conclusive tasks that must be achieved.
        Always include a call to super().
        """
        self.processor_status = ProcessorStatus.Finish
        if self.looping_status == LoopingStatus.Abort:
            self.processor_exit_status = ProcessorExitStatus.Aborted
        self.print_process_statistics()

    def print_process_statistics(self) -> None:
        """
        Print the process statistics.

        A utility method to display the fastest, the slowest and the average timing required to process on a single
        item. This is particularly useful when the looping processor is part of a ProcessorList.
        """
        if len(self._process_durations):
            log.info('[cyan] Processed %s items.' % len(self._process_durations))
            log.info(
                '[cyan] Fastest item process duration: %s '
                % pretty_format_duration(min(self._process_durations), n_digits=3)
            )
            log.info(
                '[cyan] Slowest item process duration: %s '
                % pretty_format_duration(max(self._process_durations), n_digits=3)
            )
            log.info(
                '[cyan] Average item process duration: %s '
                % pretty_format_duration((sum(self._process_durations) / len(self._process_durations)), n_digits=3)
            )
            log.info(
                '[cyan] Total process duration: %s' % pretty_format_duration(sum(self._process_durations), n_digits=3)
            )

    def _remove_orphan_files(self) -> None:
        """
        Remove orphan files.

        If a connection to the database is available, then the OrphanFile standard table is queried for all its entries,
        and all the files are then removed.

        The user can turn off this behaviour by switching the :attr:`~mafw.processor.Processor.remove_orphan_files` to False.

        """
        if self._database is None or self.remove_orphan_files is False:
            # no database connection or no wish to remove orphan files, it does not make sense to continue
            return

        # noinspection PyPep8Naming
        OrphanFile = cast(PeeweeModelWithMeta, standard_tables['OrphanFile'])

        if OrphanFile._meta.table_name not in self._database.get_tables():
            log.warning('OrphanFile table not found in DB. Please verify database integrity')
            return

        orphan_files = OrphanFile.select().execute()
        if len(orphan_files) != 0:
            msg = f'[yellow]Pruning orphan files ({len(orphan_files)})...'
            log.info(msg)
            for orphan in orphan_files:
                # filenames is a list of files:
                for f in orphan.filenames:
                    f.unlink(missing_ok=True)
            OrphanFile.delete().execute()


class ProcessorList(list[Union['Processor', 'ProcessorList']]):
    """
    A list like collection of processors.

    ProcessorList is a subclass of list containing only Processor subclasses or other ProcessorList.

    An attempt to add an element that is not a Processor or a ProcessorList will raise a TypeError.

    Along with an iterable of processors, a new processor list can be built using the following parameters.
    """

    def __init__(
        self,
        *args: Processor | ProcessorList,
        name: str | None = None,
        description: str | None = None,
        timer: Timer | None = None,
        timer_params: dict[str, Any] | None = None,
        user_interface: UserInterfaceBase | None = None,
        database: Database | None = None,
        database_conf: dict[str, Any] | None = None,
    ):
        """
        Constructor parameters:

        :param name: The name of the processor list. Defaults to ProcessorList.
        :type name: str, Optional
        :param description: An optional short description. Default to ProcessorList.
        :type description: str, Optional
        :param timer: The timer object. If None is provided, a new one will be created. Defaults to None.
        :type timer: Timer, Optional
        :param timer_params: A dictionary of parameter to build the timer object. Defaults to None.
        :type timer_params: dict, Optional
        :param user_interface: A user interface. Defaults to None
        :type user_interface: UserInterfaceBase, Optional
        :param database: A database instance. Defaults to None.
        :type database: Database, Optional
        :param database_conf: Configuration for the database. Default to None.
        :type database_conf: dict, Optional
        """

        # validate_items takes a tuple of processors, that's why we don't unpack args.
        super().__init__(self.validate_items(args))
        self._name = name or self.__class__.__name__
        self.description = description or self._name

        self.timer = timer
        self.timer_params = timer_params or {}
        self._user_interface = user_interface or ConsoleInterface()

        self._resource_stack: contextlib.ExitStack
        self._processor_exit_status: ProcessorExitStatus = ProcessorExitStatus.Successful

        # database stuff
        self._database: peewee.Database | None = database
        self._database_conf: dict[str, Any] | None = validate_database_conf(database_conf)

    def __setitem__(  # type: ignore[override]
        self,
        __index: SupportsIndex,
        __object: Processor | ProcessorList,
    ) -> None:
        super().__setitem__(__index, self.validate_item(__object))

    def insert(self, __index: SupportsIndex, __object: Processor | ProcessorList) -> None:
        """Adds a new processor at the specified index."""
        super().insert(__index, self.validate_item(__object))

    def append(self, __object: Processor | ProcessorList) -> None:
        """Appends a new processor at the end of the list."""
        super().append(self.validate_item(__object))

    def extend(self, __iterable: Iterable[Processor | ProcessorList]) -> None:
        """Extends the processor list with a list of processors."""
        if isinstance(__iterable, type(self)):
            super().extend(__iterable)
        else:
            super().extend([self.validate_item(item) for item in __iterable])

    @staticmethod
    def validate_item(item: Processor | ProcessorList) -> Processor | ProcessorList:
        """Validates the item being added."""
        if isinstance(item, Processor):
            item.local_resource_acquisition = False
            return item
        elif isinstance(item, ProcessorList):
            item.timer_params = dict(suppress_message=True)
            return item
        else:
            raise TypeError(f'Expected Processor or ProcessorList, got {type(item).__name__}')

    @staticmethod
    def validate_items(items: tuple[Processor | ProcessorList, ...] = ()) -> tuple[Processor | ProcessorList, ...]:
        """Validates a tuple of items being added."""
        if not items:
            return tuple()
        return tuple([ProcessorList.validate_item(item) for item in items if item is not None])

    @property
    def name(self) -> str:
        """
        The name of the processor list

        :return: The name of the processor list
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def processor_exit_status(self) -> ProcessorExitStatus:
        """
        The processor exit status.

        It refers to the whole processor list execution.
        """
        return self._processor_exit_status

    @processor_exit_status.setter
    def processor_exit_status(self, status: ProcessorExitStatus) -> None:
        self._processor_exit_status = status

    @property
    def database(self) -> peewee.Database:
        """
        Returns the database instance

        :return: A database instance
        :raises MissingDatabase: if a database connection is missing.
        """
        if self._database is None:
            raise MissingDatabase('Database connection not initialized')
        return self._database

    def execute(self) -> ProcessorExitStatus:
        """
        Execute the list of processors.

        Similarly to the :class:`Processor`, ProcessorList can be executed. In simple words, the execute
        method of each processor in the list is called exactly in the same sequence as they were added.
        """
        with contextlib.ExitStack() as self._resource_stack:
            self.acquire_resources()
            self._user_interface.create_task(self.name, self.description, completed=0, increment=0, total=len(self))
            for i, item in enumerate(self):
                if isinstance(item, Processor):
                    log.info('Executing [red]%s[/red] processor' % item.name)
                else:
                    log.info('Executing [red]%s[/red] processor list' % item.name)
                self.distribute_resources(item)
                item.execute()
                self._user_interface.update_task(self.name, i, 1, len(self))
                self._processor_exit_status = item.processor_exit_status
                if self._processor_exit_status == ProcessorExitStatus.Aborted:
                    msg = 'Processor %s caused the processor list to abort' % item.name
                    log.error(msg)
                    raise AbortProcessorException(msg)
            self._user_interface.update_task(self.name, completed=len(self), total=len(self))
        return self._processor_exit_status

    def acquire_resources(self) -> None:
        """Acquires external resources."""
        # The strategy is similar to the one for processor. if we do get resources already active (not None) then we use
        # them, otherwise, we create them and we add them to the resource stack.
        if self.timer is None:
            self.timer = self._resource_stack.enter_context(Timer(**self.timer_params))
        self._resource_stack.enter_context(self._user_interface)
        if self._database is None and self._database_conf is None:
            # no database, nor  configuration.
            # we cannot do anything
            pass
        elif self._database is None and self._database_conf is not None:
            # no db, but we got a configuration.
            # we can make a db
            if 'DBConfiguration' in self._database_conf:
                conf = self._database_conf['DBConfiguration']  # type1
            else:
                conf = self._database_conf  # type2

            # guess the database type from the URL
            protocol = extract_protocol(conf.get('URL'))

            # build the connection parameter
            # in case of sqlite, we add the pragmas group as well
            connection_parameters = {}
            if protocol == 'sqlite':
                connection_parameters['pragmas'] = conf.get('pragmas', {})
            for key, value in conf.items():
                if key not in ['URL', 'pragmas']:
                    connection_parameters[key] = value

            self._database = connect(conf.get('URL'), **connection_parameters)  # type: ignore # peewee is not returning a DB
            try:
                self._database.connect()
                self._resource_stack.callback(self._database.close)
            except peewee.OperationalError as e:
                log.critical('Unable to connect to %s', self._database_conf.get('URL'))
                raise e
            database_proxy.initialize(self._database)
            self.database.create_tables(standard_tables.values())
            for table in standard_tables.values():
                table.init()
        else:  # equiv to if self._database is not None:
            # we got a database, so very likely we are inside a processor list
            # the connection has been already set and the initialisation as well.
            # nothing else to do here.
            pass

    def distribute_resources(self, processor: Processor | Self) -> None:
        """Distributes the external resources to the items in the list."""
        processor.timer = self.timer
        processor._user_interface = self._user_interface
        processor._database = self._database



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Module implements the abstract base interface to a processor to generate plots.

This abstract interface is needed because MAFw does not force the user to select a specific plot and data manipulation
library.

The basic idea is to have a :class:`basic processor class <.GenericPlotter>` featuring a modified
:meth:`~.GenericPlotter.process` method where a skeleton of the standard operations required to generate a graphical
representation of a dataset is provided.

The user has the possibility to compose the :class:`~.GenericPlotter` by mixing it with one :class:`~.DataRetriever`
and a :class:`~.FigurePlotter`.

For a specific implementation based on :link:`seaborn`, please refer to :mod:`.sns_plotter`.
"""

import logging
import typing
from abc import ABC, abstractmethod
from typing import Any, Protocol

import peewee

from mafw.db.std_tables import PlotterOutput, TriggerDisabler
from mafw.enumerators import LoopingStatus
from mafw.processor import Processor, ProcessorMeta
from mafw.tools.file_tools import file_checksum

log = logging.getLogger(__name__)


class PlotterMeta(type(Protocol), ProcessorMeta):  # type: ignore[misc]
    """Metaclass for the plotter mixed classes"""

    pass


class DataRetriever(ABC):
    """Base mixin class to retrieve a data frame from an external source"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # leave it here, otherwise the Protocol init will not call the main class init.
        # not sure why this is happening, but it costs nothing to have it here.

        """The dataframe instance. It will be filled for the main class"""
        super().__init__(*args, **kwargs)

    @abstractmethod
    def get_data_frame(self) -> None:
        """The mixin implementation of the shared method with the base class"""
        pass  # pragma: no cover

    @abstractmethod
    def patch_data_frame(self) -> None:
        """The mixin implementation of the shared method with the base class"""
        pass  # pragma: no cover

    @abstractmethod
    def _attributes_valid(self) -> bool:
        pass  # pragma: no cover


class FigurePlotter(ABC):
    @abstractmethod
    def plot(self) -> None:
        pass  # pragma: no cover

    @abstractmethod
    def _attributes_valid(self) -> bool:
        pass  # pragma: no cover


class GenericPlotter(Processor, metaclass=PlotterMeta):
    """
    The Generic Plotter processor.

    This is a subclass of a Processor with advanced functionality to fetch data in the form of a dataframe and to
    produce plots. When mentioning dataframe in the context of the generic plotter, we do not have in mind any
    specific dataframe implementation.

    The GenericPlotter is actually a kind of abstract class: since MAFw is not forcing you to use any specific
    plotting and data manipulation library, you need to subclass the GenericPlotter in your code, be sure that the
    required dependencies are available for import and use it as a normal processor.

    If you are ok with using :link:`seaborn` (with :link:`matplotlib` as a graphical backend and :link:`pandas` for
    data storage and manipulation), then be sure to install mafw with the optional feature `seaborn` (``pip install
    mafw[seaborn]``) and have a look at the :mod:`~.sns_plotter` for an already prepared implementation of a Plotter.

    The key difference with respect to a normal processor is its :meth:`.process` method that has been already
    implemented as follows:

    .. literalinclude:: ../../../src/mafw/processor_library/abstract_plotter.py
        :pyobject: GenericPlotter.process
        :dedent:

    This actually means that when you are subclassing a GenericPlotter you do not have to implement the process method
    as you would do for a normal Processor, but you will have to implement the following methods:

        * :meth:`~.in_loop_customization`.

            The processor execution workflow (LoopType) can be any of the available, so
            actually the process method might be invoked only once, or multiple times inside a loop structure
            (for or while).
            If the execution is cyclic, then you may want to have the possibility to do some customisation for each
            iteration, for example, changing the plot title, or modifying the data selection, or the filename where the
            plots will be saved.

            You can use this method also in case of a single loop processor, in this case you will not have access to
            the loop parameters.

        * :meth:`~.get_data_frame`.

            This method has the task to get the data to be plotted. Since it is an almost abstract class, you need to

        * :meth:`~.patch_data_frame`.

            A convenient method to apply data frame manipulation to the data just retrieved. A typical use case is for
            conversion of unit of measurement. Imagine you saved the data in the S.I. units, but for the visualization
            you prefer to use practical units, so you can subclass this method to add a new column containing the same
            converted values of the original one.

        * :meth:`~.slice_data_frame`.

            Slicing a dataframe is similar as applying a where clause in a SQL query. Implement this method to select
            which row should be used in the generation of your plot.

        * :meth:`~.group_and_aggregate_data_frame`.

            In this method, you can manipulate your data frame to perform row grouping and aggregation.

        * :meth:`~.is_data_frame_empty`.

            A simple method to test if the dataframe contains any data to be plotted. In fact, after the slicing, grouping
            and aggregation operations, it is possible that the dataframe is now left without any row. In this case,
            it makes no sense to waste time in plotting an empty graph.

        * :meth:`~.plot`.

            This method is where the actual plotting occurs.

        * :meth:`~.customize_plot`.

            This method can be optionally used to customize the appearance of the facet grid produced by the
            :meth:`~plot` method. It is particularly useful when the user is mixing this class with one of the
            :class:`~.FigurePlotter` mixin, thus not having direct access to the plot method.

        * :meth:`~.save`.

            This method is where the produced plot is saved in a file. Remember to append the output file name to the
            :attr:`list of produced outputs <.output_filename_list>` so that the :meth:`~._update_plotter_db` method
            will automatically store this file in the database during the :meth:`~.finish` execution.

        * :meth:`~.update_db`.

            If the user wants to update a specific table in the database, they can use this method.

            It is worth reminding that all plotters are saving all generated files in the standard table PlotterOutput.
            This is automatically done by the :meth:`~._update_plotter_db` method that is called in the
            :meth:`~.finish` method.

    """

    @typing.no_type_check
    def is_output_existing(self) -> bool:
        """
        Check for plotter output existence.

        Generally, plotter subclasses do not have a real output that can be saved to a database. This class is meant to
        generate one or more graphical output files.

        One of the biggest advantages of having the output of a processor stored in the database is the ability to
        conditionally execute the processor if, and only if, the output is missing or changed.

        In order to allow also plotter processor to benefit from this feature, a :class:`dedicated table
        <.PlotterOutput>` is available among the :ref:`standard tables <std_tables>`.

        If a connection to the database is provided, then this method is invoked at the beginning of the
        :meth:`~.process` and a select query over the :class:`~.PlotterOutput` model is executed filtering by
        processor name. All files in the filename lists are checked for existence and also the checksum is verified.

        This method will return True, if the output of the processor is already existing and valid, False, otherwise.

        :return: True if the processor output exists and it is valid.
        :rtype: bool
        """
        if self._database is None:
            # no active database connection. it makes no sense to continue. inform the user and return
            log.warning('No database connection available. Impossible to check for existing output')
            return False

        try:
            query = PlotterOutput.get(PlotterOutput.plotter_name == self.name)
            # check if all files exist:
            if not all([f.exists() for f in query.filename_list]):
                # at least one file is missing.
                # delete the whole row and continue
                with TriggerDisabler(trigger_type_id=4):
                    PlotterOutput.delete().where(PlotterOutput.plotter_name == self.name).execute()

                return False
            else:
                # all files exist.
                # check that they are still actual
                if query.checksum != file_checksum(query.filename_list):
                    # at least one file is changed.
                    # delete the whole row and continue
                    with TriggerDisabler(trigger_type_id=4):
                        PlotterOutput.delete().where(PlotterOutput.plotter_name == self.name).execute()
                    return False
                else:
                    # all files exit and the checksum is the same.
                    # we stop it here
                    return True

        except peewee.DoesNotExist:
            # no output for this plotter processor found in the DB.
            return False

    def process(self) -> None:
        """
        Process method overload.

        In the case of a plotter subclass, the process method is already implemented and the user should not overload
        it. On the contrary, the user must overload the other implementation methods described in the general
        :class:`class description <.SNSPlotter>`.
        """
        if self.filter_register.new_only:
            if self.is_output_existing():
                return

        self.in_loop_customization()
        self.get_data_frame()
        self.patch_data_frame()
        self.slice_data_frame()
        self.group_and_aggregate_data_frame()
        if not self.is_data_frame_empty():
            self.plot()
            self.customize_plot()
            self.save()
            self.update_db()

    def is_data_frame_empty(self) -> bool:
        """Check if the data frame is empty"""
        return False

    def in_loop_customization(self) -> None:
        """
        Customize the parameters for the output or input data for each execution iteration.
        """
        pass

    def get_data_frame(self) -> None:
        """
        Get the data frame with the data to be plotted.

        This method can be either implemented in the SNSPlotter subclass or via a :class:`.DataRetriever` mixin
        class.
        """
        # it must be overloaded.
        pass

    def format_progress_message(self) -> None:
        self.progress_message = f'{self.name} is working'

    def plot(self) -> None:
        """
        The plot method.

        This is where the user has to implement the real plot generation
        """
        pass

    def customize_plot(self) -> None:
        """
        The customize plot method.

        The user can overload this method to customize the output produced by the :meth:`~.plot` method, like, for
        example, adding meaningful axis titles, changing format, and so on.

        As usual, it is possible to use the :attr:`~.item`, :attr:`~.i_item` and :attr:`~.n_item` to access the loop
        parameters.
        """
        pass

    def save(self) -> None:
        """
        The save method.

        This is where the user has to implement the saving of the plot on disc.
        """
        pass

    def update_db(self) -> None:
        """
        The update database method.

        This is where the user has to implement the optional update of the database.

        .. seealso:

            The plotter output table is automatically update by :meth:`~._update_plotter_db`.
        """
        pass

    def slice_data_frame(self) -> None:
        pass

    def group_and_aggregate_data_frame(self) -> None:
        pass

    def finish(self) -> None:
        if self.looping_status == LoopingStatus.Continue:
            self._update_plotter_db()  # type: ignore[no-untyped-call]
        super().finish()

    def patch_data_frame(self) -> None:
        """
        Modify the data frame

        This method can be used to perform operation on the data frame, like adding new columns.
        It can be either implemented in the plotter processor subclasses or via a mixin class.
        """
        pass

    @typing.no_type_check
    def _update_plotter_db(self) -> None:
        """
        Updates the Plotter DB.

        A plotter subclass primarily generates plots as output in most cases, which means that no additional information needs to be stored in the database. This is sufficient to prevent unnecessary execution of the processor when it is not required.

        This method is actually protected against execution without a valid database instance.
        """
        if self._database is None:
            # there is no active database connection. No need to continue. Inform the user and continue
            log.warning('No database connection available. Impossible to update the plotter output')
            return

        if len(self.output_filename_list) == 0:
            # there is no need to make an entry because there are no saved file
            return

        PlotterOutput.std_upsert(
            {
                'plotter_name': self.name,
                'filename_list': self.output_filename_list,
                'checksum': self.output_filename_list,
            }
        ).execute()



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
The module provides some examples on how to use ProcessorList to combine several processors.
"""


def run_simple_processor_list():
    """Simplest way to run several processors in a go."""
    from mafw.examples.sum_processor import AccumulatorProcessor, GaussAdder
    from mafw.processor import ProcessorList

    # create the list. name and description are optional
    new_list = ProcessorList(name='AddingProcessor', description='Summing up numbers')

    # append the processors. you can pass parameters to the processors in the standard way
    max_value = 120
    new_list.append(AccumulatorProcessor(last_value=max_value))
    new_list.append(GaussAdder(last_value=max_value))

    # execute the list. This will execute all the processors in the list
    new_list.execute()

    # you can access single processors in the list, in the standard way.
    # remember that the ProcessorList is actually a list!
    assert new_list[0].accumulated_value == new_list[1].sum_value


def run_processor_list_with_loop_modifier():
    """Example on deal with processors inside a processor list changing the loop structure.

    In this example there are two processors, one that will run until the end and the other that will set the looping
    status to abort half way. The user can see what happens when the :class:`~mafw.processor.ProcessorList` is executed.
    """
    import time

    from mafw.enumerators import LoopingStatus, ProcessorExitStatus, ProcessorStatus
    from mafw.mafw_errors import AbortProcessorException
    from mafw.processor import ActiveParameter, Processor, ProcessorList

    class GoodProcessor(Processor):
        n_loop = ActiveParameter('n_loop', default=100, help_doc='The n of the loop')
        sleep_time = ActiveParameter('sleep_time', default=0.01, help_doc='So much work')

        def get_items(self) -> list[int]:
            return list(range(self.n_loop))

        def process(self):
            # pretend to do something, but actually sleep
            time.sleep(self.sleep_time)

        def finish(self):
            super().finish()
            print(f'{self.name} just finished with status: {self.processor_exit_status.name}')

    class BadProcessor(Processor):
        n_loop = ActiveParameter('n_loop', default=100, help_doc='The n of the loop')
        sleep_time = ActiveParameter('sleep_time', default=0.01, help_doc='So much work')
        im_bad = ActiveParameter('im_bad', default=50, help_doc='I will crash it!')

        def get_items(self) -> list[int]:
            return list(range(self.n_loop))

        def process(self):
            if self.item == self.im_bad:
                self.looping_status = LoopingStatus.Abort
                return
            # let me do my job
            time.sleep(self.sleep_time)

        def finish(self):
            super().finish()
            print(f'{self.name} just finished with status: {self.processor_exit_status.name}')

    proc_list = ProcessorList(name='with exception')
    proc_list.extend([GoodProcessor(), BadProcessor(), GoodProcessor()])
    try:
        proc_list.execute()
    except AbortProcessorException:
        print('I know you were a bad guy')
    assert proc_list.processor_exit_status == ProcessorExitStatus.Aborted
    assert proc_list[0].processor_exit_status == ProcessorExitStatus.Successful
    assert proc_list[1].processor_exit_status == ProcessorExitStatus.Aborted
    assert proc_list[2].processor_status == ProcessorStatus.Init



"""
An abstract generic user interface.

The module provides a generic user interface that can be implemented to allow MAFw to communicate with different
user interfaces.

MAFw is designed to operate seamlessly without a user interface; however, users often appreciate the added benefit of communication between the process execution and themselves.

There are several different interfaces and different interface types (Command Line, Textual, Graphical...) and
everyone has its own preferences. In order to be as generic as possible, MAFw is allowing for an abstract
interface layer so that the user can either decide to use one of the few coming with MAFw or to implement the
interface to their favorite interface.
"""

from __future__ import annotations

from types import TracebackType
from typing import Any, Self

from mafw.enumerators import ProcessorStatus


class UserInterfaceMeta(type):
    """
    A metaclass used for the creation of user interface
    """

    __required_members__ = ('create_task', 'update_task', 'display_progress_message', 'change_of_processor_status')
    __required_callable__ = ('create_task', 'update_task', 'display_progress_message', 'change_of_processor_status')

    def __instancecheck__(cls, instance: Any) -> bool:
        return cls.__subclasscheck__(type(instance))

    def __subclasscheck__(cls, subclass: Any) -> bool:
        members = all([hasattr(subclass, m) for m in cls.__required_members__])
        if not members:
            return False
        callables = all([callable(getattr(subclass, c)) for c in cls.__required_callable__])
        return callables


class UserInterfaceBase(metaclass=UserInterfaceMeta):
    """The abstract base user interface class."""

    always_display_progress_message = 10
    """
    Threshold for displaying progress messages.
    
    If the total number of events is below this value, then the progress message is always displayed, otherwise 
    follow the standard update frequency.
    """
    name = 'base'
    """The name of the interface"""

    def create_task(
        self,
        task_name: str,
        task_description: str = '',
        completed: int = 0,
        increment: int | None = None,
        total: int | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Create a new task.

        :param task_name: A unique identifier for the task. You cannot have more than 1 task with the same name in
            the whole execution. If you want to use the processor name, it is recommended to use the
            :attr:`~mafw.processor.Processor.unique_name`.
        :type task_name: str
        :param task_description: A short description for the task. Defaults to None.
        :type task_description: str, Optional
        :param completed: The amount of task already completed. Defaults to None.
        :type completed: int, Optional
        :param increment: How much of the task has been done since last update. Defaults to None.
        :type increment: int, Optional
        :param total: The total amount of task. Defaults to None.
        :type total: int, Optional
        """
        pass

    def update_task(
        self, task_name: str, completed: int = 0, increment: int | None = None, total: int | None = None, **kwargs: Any
    ) -> None:
        """
        Update an existing task.

        :param task_name: A unique identifier for the task. You cannot have more than 1 task with the same name in
            the whole execution. If you want to use the processor name, it is recommended to use the
            :attr:`~mafw.processor.Processor.unique_name`.
        :type task_name: str
        :param completed: The amount of task already completed. Defaults to None.
        :type completed: int, Optional
        :param increment: How much of the task has been done since last update. Defaults to None.
        :type increment: int, Optional
        :param total: The total amount of task. Defaults to None.
        :type total: int, Optional
        """
        pass

    def __enter__(self) -> Self:
        """Context enter dunder."""
        return self

    def __exit__(
        self, type_: type[BaseException] | None, value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        """
        Context exit dunder.

        :param type_: Exception type.
        :param value: Exception value.
        :param traceback: Exception trace back.
        """
        pass

    def display_progress_message(self, message: str, i_item: int, n_item: int | None, frequency: float) -> None:
        """
        Display a message during the process execution.

        :param message: The message to be displayed.
        :type message: str
        :param i_item: The current item enumerator.
        :type i_item: int
        :param n_item: The total number of items or None for an indeterminate progress (while loop).
        :type n_item: int | None
        :param frequency: How often (in percentage of n_item) to display the message.
        :type frequency: float
        """
        pass

    def _is_time_to_display_lopping_message(self, i_item: int, n_item: int | None, frequency: float) -> bool:
        if n_item is None:
            # let's print the message everytime i_item is a multiple of 10.
            return i_item % 10 == 0

        if n_item > self.always_display_progress_message:
            always = False
        else:
            always = True
        if always:
            do_display = True
        else:
            mod = max([round(frequency * n_item), 1])
            do_display = i_item == 0 or i_item % mod == 0 or i_item == n_item - 1
        return do_display

    def change_of_processor_status(
        self, processor_name: str, old_status: ProcessorStatus, new_status: ProcessorStatus
    ) -> None:
        pass



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Unit tests for the rich_user_interface module.

This module contains comprehensive tests for the RichInterface class,
ensuring good coverage of all methods and edge cases.
"""

from unittest.mock import Mock, patch

import pytest
from rich.progress import Progress, TaskID

from mafw.enumerators import ProcessorStatus
from mafw.ui.rich_user_interface import RichInterface


class TestRichInterfaceInitialization:
    """Test class for RichInterface initialization and setup."""

    def test_default_initialization(self):
        """Test RichInterface initialization with default parameters."""
        interface = RichInterface()

        assert interface.name == 'rich'
        assert isinstance(interface.progress, Progress)
        assert isinstance(interface.task_dict, dict)
        assert len(interface.task_dict) == 0

    def test_initialization_with_custom_progress_kws(self):
        """Test RichInterface initialization with custom progress keywords."""
        custom_kws = {'auto_refresh': False, 'expand': False, 'refresh_per_second': 5}

        with patch('mafw.ui.rich_user_interface.Progress') as mock_progress:
            RichInterface(progress_kws=custom_kws)

            # Verify Progress was called with custom keywords
            mock_progress.assert_called_once()
            call_args = mock_progress.call_args
            assert 'auto_refresh' in call_args.kwargs
            assert call_args.kwargs['auto_refresh'] is False
            assert call_args.kwargs['expand'] is False
            assert call_args.kwargs['refresh_per_second'] == 5

    def test_initialization_with_none_progress_kws(self):
        """Test RichInterface initialization when progress_kws is explicitly None."""
        with patch('mafw.ui.rich_user_interface.Progress') as mock_progress:
            RichInterface(progress_kws=None)

            # Should use default kwargs
            call_args = mock_progress.call_args
            assert 'auto_refresh' in call_args.kwargs
            assert call_args.kwargs['auto_refresh'] is True
            assert call_args.kwargs['expand'] is True


class TestRichInterfaceContextManager:
    """Test class for RichInterface context manager functionality."""

    def test_context_manager_enter(self):
        """Test __enter__ method starts progress and returns self."""
        interface = RichInterface()
        interface.progress = Mock()

        result = interface.__enter__()

        interface.progress.start.assert_called_once()
        assert result is interface

    def test_context_manager_exit(self):
        """Test __exit__ method stops progress."""
        interface = RichInterface()
        interface.progress = Mock()

        interface.__exit__(None, None, None)

        interface.progress.stop.assert_called_once()

    @pytest.mark.parametrize(
        'exception_type,exception_value,traceback',
        [
            (ValueError, ValueError('test error'), None),
            (RuntimeError, RuntimeError('runtime error'), Mock()),
            (None, None, None),
        ],
    )
    def test_context_manager_exit_with_exceptions(self, exception_type, exception_value, traceback):
        """Test __exit__ method with various exception scenarios."""
        interface = RichInterface()
        interface.progress = Mock()

        interface.__exit__(exception_type, exception_value, traceback)

        interface.progress.stop.assert_called_once()

    def test_context_manager_full_flow(self):
        """Test complete context manager flow."""
        interface = RichInterface()
        interface.progress = Mock()

        with interface as ctx:
            assert ctx is interface
            interface.progress.start.assert_called_once()

        interface.progress.stop.assert_called_once()


class TestRichInterfaceTaskManagement:
    """Test class for RichInterface task creation and updates."""

    def test_create_task_basic(self):
        """Test basic task creation."""
        interface = RichInterface()
        interface.progress = Mock()
        mock_task_id = TaskID(1)
        interface.progress.add_task.return_value = mock_task_id

        interface.create_task('test_task', 'Test Description')

        interface.progress.add_task.assert_called_once_with('Test Description', total=None, completed=0, increment=None)
        assert interface.task_dict['test_task'] == mock_task_id

    @pytest.mark.parametrize(
        'task_name,task_description,completed,increment,total',
        [
            ('task1', 'Description 1', 0, None, None),
            ('task2', 'Description 2', 10, 5, 100),
            ('task3', '', 25, None, 50),
            ('task4', 'Long description with details', 0, 1, None),
        ],
    )
    def test_create_task_with_parameters(self, task_name, task_description, completed, increment, total):
        """Test task creation with various parameter combinations."""
        interface = RichInterface()
        interface.progress = Mock()
        mock_task_id = TaskID(1)
        interface.progress.add_task.return_value = mock_task_id

        interface.create_task(task_name, task_description, completed, increment, total)

        interface.progress.add_task.assert_called_once_with(
            task_description, total=total, completed=completed, increment=increment
        )
        assert interface.task_dict[task_name] == mock_task_id

    def test_create_task_with_kwargs(self):
        """Test task creation with additional keyword arguments."""
        interface = RichInterface()
        interface.progress = Mock()
        mock_task_id = TaskID(1)
        interface.progress.add_task.return_value = mock_task_id

        interface.create_task('test_task', 'Test Description', custom_arg='value', another_arg=42)

        # Should still work even with extra kwargs (they're ignored)
        interface.progress.add_task.assert_called_once_with('Test Description', total=None, completed=0, increment=None)

    @patch('mafw.ui.rich_user_interface.log')
    def test_create_task_duplicate_name_warning(self, mock_log):
        """Test warning when creating task with duplicate name."""
        interface = RichInterface()
        interface.progress = Mock()
        mock_task_id1 = TaskID(1)
        mock_task_id2 = TaskID(2)
        interface.progress.add_task.side_effect = [mock_task_id1, mock_task_id2]

        # Create first task
        interface.create_task('duplicate_task', 'First Description')

        # Create second task with same name
        interface.create_task('duplicate_task', 'Second Description')

        # Check warnings were logged
        assert mock_log.warning.call_count == 2
        mock_log.warning.assert_any_call(
            'A task with this name (%s) already exists. Replacing it with the new one.' % 'duplicate_task'
        )
        mock_log.warning.assert_any_call('Be sure to use unique names.')

        # Check task was replaced
        assert interface.task_dict['duplicate_task'] == mock_task_id2

    def test_update_task_basic(self):
        """Test basic task update."""
        interface = RichInterface()
        interface.progress = Mock()
        mock_task_id = TaskID(1)
        interface.task_dict['existing_task'] = mock_task_id

        interface.update_task('existing_task', completed=50)

        interface.progress.update.assert_called_once_with(
            mock_task_id,
            completed=50,
            advance=None,
            total=None,
            visible=True,  # completed != total (None)
        )

    @pytest.mark.parametrize(
        'completed,increment,total,expected_visible',
        [
            (0, None, None, True),
            (50, 10, 100, True),
            (100, None, 100, False),  # completed == total
            (0, 5, 10, True),
            (10, None, 10, False),  # completed == total
        ],
    )
    def test_update_task_visibility(self, completed, increment, total, expected_visible):
        """Test task update visibility based on completion status."""
        interface = RichInterface()
        interface.progress = Mock()
        mock_task_id = TaskID(1)
        interface.task_dict['test_task'] = mock_task_id

        interface.update_task('test_task', completed=completed, increment=increment, total=total)

        interface.progress.update.assert_called_once_with(
            mock_task_id, completed=completed, advance=increment, total=total, visible=expected_visible
        )

    @patch('mafw.ui.rich_user_interface.log')
    def test_update_task_nonexistent(self, mock_log):
        """Test updating a non-existent task logs warning and returns early."""
        interface = RichInterface()
        interface.progress = Mock()

        interface.update_task('nonexistent_task', completed=50)

        # Check warnings were logged
        assert mock_log.warning.call_count == 2
        mock_log.warning.assert_any_call('A task with this name (%s) does not exist.' % 'nonexistent_task')
        mock_log.warning.assert_any_call('Skipping updates')

        # Check progress.update was not called
        interface.progress.update.assert_not_called()

    def test_update_task_with_kwargs(self):
        """Test task update with additional keyword arguments."""
        interface = RichInterface()
        interface.progress = Mock()
        mock_task_id = TaskID(1)
        interface.task_dict['test_task'] = mock_task_id

        interface.update_task('test_task', completed=25, increment=5, total=100, extra_arg='ignored')

        interface.progress.update.assert_called_once_with(
            mock_task_id, completed=25, advance=5, total=100, visible=True
        )


class TestRichInterfaceProgressMessages:
    """Test class for RichInterface progress message display."""

    @patch('mafw.ui.rich_user_interface.log')
    def test_display_progress_message_with_n_item(self, mock_log):
        """Test progress message display with known total items."""
        interface = RichInterface()

        # Mock the _is_time_to_display_lopping_message method to always return True
        interface._is_time_to_display_lopping_message = Mock(return_value=True)

        interface.display_progress_message('Processing item', 4, 100, 0.1)

        # Check that log.info was called with properly formatted message
        expected_message = '[  5/100] Processing item'
        mock_log.info.assert_called_once_with(expected_message)

    @patch('mafw.ui.rich_user_interface.log')
    def test_display_progress_message_without_n_item(self, mock_log):
        """Test progress message display without known total items."""
        interface = RichInterface()
        interface._is_time_to_display_lopping_message = Mock(return_value=True)

        interface.display_progress_message('Processing item', 4, None, 0.1)

        expected_message = '[   5/1000] Processing item'
        mock_log.info.assert_called_once_with(expected_message)

    @patch('mafw.ui.rich_user_interface.log')
    def test_display_progress_message_large_i_item(self, mock_log):
        """Test progress message display with large i_item value."""
        interface = RichInterface()
        interface._is_time_to_display_lopping_message = Mock(return_value=True)

        interface.display_progress_message('Processing item', 1500, None, 0.1)

        # When i_item > 1000, n_item should be i_item
        expected_message = '[1501/1500] Processing item'
        mock_log.info.assert_called_once_with(expected_message)

    @patch('mafw.ui.rich_user_interface.log')
    def test_display_progress_message_not_time_to_display(self, mock_log):
        """Test progress message when it's not time to display."""
        interface = RichInterface()
        interface._is_time_to_display_lopping_message = Mock(return_value=False)

        interface.display_progress_message('Processing item', 4, 100, 0.1)

        # Should not log anything
        mock_log.info.assert_not_called()

    @patch('mafw.ui.rich_user_interface.log')
    @pytest.mark.parametrize(
        'i_item,n_item,expected_width,expected_counter',
        [
            (0, 10, 2, '[ 1/10] '),
            (9, 100, 3, '[ 10/100] '),
            (99, 1000, 4, '[ 100/1000] '),
            (1999, 10000, 5, '[ 2000/10000] '),
        ],
    )
    def test_display_progress_message_formatting(self, mock_log, i_item, n_item, expected_width, expected_counter):
        """Test progress message formatting with different widths."""
        interface = RichInterface()
        interface._is_time_to_display_lopping_message = Mock(return_value=True)

        interface.display_progress_message('Test message', i_item, n_item, 0.1)

        expected_message = expected_counter + 'Test message'
        mock_log.info.assert_called_once_with(expected_message)


class TestRichInterfaceStatusChanges:
    """Test class for RichInterface processor status changes."""

    @patch('mafw.ui.rich_user_interface.log')
    @pytest.mark.parametrize(
        'processor_name,old_status,new_status',
        [
            ('processor1', ProcessorStatus.Unknown, ProcessorStatus.Init),
            ('processor2', ProcessorStatus.Init, ProcessorStatus.Run),
            ('processor3', ProcessorStatus.Run, ProcessorStatus.Finish),
            ('long_processor_name', ProcessorStatus.Finish, ProcessorStatus.Unknown),
        ],
    )
    def test_change_of_processor_status(self, mock_log, processor_name, old_status, new_status):
        """Test processor status change logging."""
        interface = RichInterface()

        interface.change_of_processor_status(processor_name, old_status, new_status)

        expected_message = f'[red]{processor_name}[/red] is [bold]{new_status}[/bold]'
        mock_log.debug.assert_called_once_with(expected_message)

    @patch('mafw.ui.rich_user_interface.log')
    def test_change_of_processor_status_with_special_characters(self, mock_log):
        """Test processor status change with special characters in name."""
        interface = RichInterface()
        processor_name = 'processor-with-dashes_and_underscores.123'

        interface.change_of_processor_status(processor_name, ProcessorStatus.Init, ProcessorStatus.Run)

        expected_message = f'[red]{processor_name}[/red] is [bold]{ProcessorStatus.Run}[/bold]'
        mock_log.debug.assert_called_once_with(expected_message)


@pytest.mark.integration_test
class TestRichInterfaceIntegration:
    """Integration tests for RichInterface."""

    def test_complete_workflow(self):
        """Test a complete workflow with task creation, updates, and status changes."""
        interface = RichInterface()
        interface.progress = Mock()
        mock_task_id = TaskID(1)
        interface.progress.add_task.return_value = mock_task_id

        # Test context manager
        with interface:
            # Create task
            interface.create_task('workflow_task', 'Processing workflow', total=100)

            # Update task multiple times
            interface.update_task('workflow_task', completed=25)
            interface.update_task('workflow_task', completed=50)
            interface.update_task('workflow_task', completed=100)

            # Test status change
            interface.change_of_processor_status('test_processor', ProcessorStatus.Unknown, ProcessorStatus.Init)

        # Verify all calls
        interface.progress.start.assert_called_once()
        interface.progress.add_task.assert_called_once()
        assert interface.progress.update.call_count == 3
        interface.progress.stop.assert_called_once()

    @patch('mafw.ui.rich_user_interface.log')
    def test_multiple_tasks_management(self, mock_log):
        """Test managing multiple tasks simultaneously."""
        interface = RichInterface()
        interface.progress = Mock()
        interface.progress.add_task.side_effect = [TaskID(1), TaskID(2), TaskID(3)]

        # Create multiple tasks
        tasks = [
            ('task1', 'First task', 0, None, 100),
            ('task2', 'Second task', 10, 5, 200),
            ('task3', 'Third task', 25, None, 50),
        ]

        for task_name, description, completed, increment, total in tasks:
            interface.create_task(task_name, description, completed, increment, total)

        # Update all tasks
        interface.update_task('task1', completed=50)
        interface.update_task('task2', completed=75, increment=10)
        interface.update_task('task3', completed=50, total=50)  # This should be invisible

        # Verify all tasks were created and updated
        assert len(interface.task_dict) == 3
        assert interface.progress.add_task.call_count == 3
        assert interface.progress.update.call_count == 3

        # Check the last update call for task3 (should be invisible)
        last_call = interface.progress.update.call_args_list[-1]
        assert last_call.kwargs['visible'] is False

    def test_inheritance_from_base_class(self):
        """Test that RichInterface properly inherits from UserInterfaceBase."""
        from mafw.ui.abstract_user_interface import UserInterfaceBase

        interface = RichInterface()
        assert isinstance(interface, UserInterfaceBase)

        # Test that all required methods are implemented
        required_methods = ['create_task', 'update_task', 'display_progress_message', 'change_of_processor_status']
        for method_name in required_methods:
            assert hasattr(interface, method_name)
            assert callable(getattr(interface, method_name))

    def test_progress_message_integration_with_base_class(self):
        """Test that progress message display integrates with base class logic."""
        interface = RichInterface()

        # Test that the base class method is used for determining when to display
        with patch.object(interface, '_is_time_to_display_lopping_message', return_value=True) as mock_method:
            with patch('mafw.ui.rich_user_interface.log') as mock_log:
                interface.display_progress_message('Test', 5, 100, 0.1)

                mock_method.assert_called_once_with(5, 100, 0.1)
                mock_log.info.assert_called_once()



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
The module provides tools to read / write / modify specific TOML files.
"""

import datetime
import logging
from pathlib import Path, PosixPath, WindowsPath
from typing import Any, cast

import tomlkit
from tomlkit import TOMLDocument, comment, document, item, nl, table
from tomlkit.exceptions import ConvertError
from tomlkit.items import Item, String, StringType
from tomlkit.toml_file import TOMLFile

import mafw.mafw_errors
from mafw.db.db_configurations import default_conf
from mafw.mafw_errors import InvalidSteeringFile
from mafw.processor import PassiveParameter, Processor

log = logging.getLogger(__name__)


class PathItem(String):
    """TOML item representing a Path"""

    def unwrap(self) -> Path:  # type: ignore[override] # do not know how to do it
        return Path(super().unwrap())


def path_encoder(obj: Any) -> Item:
    """Encoder for PathItem."""
    if isinstance(obj, PosixPath):
        return PathItem.from_raw(str(obj), type_=StringType.SLB, escape=False)
    elif isinstance(obj, WindowsPath):
        return PathItem.from_raw(str(obj), type_=StringType.SLL, escape=False)
    else:
        raise ConvertError


tomlkit.register_encoder(path_encoder)


def generate_steering_file(
    output_file: Path | str,
    processors: list[type[Processor] | Processor] | type[Processor] | Processor,
    database_conf: dict[str, Any] | None = None,
    db_engine: str = 'sqlite',
) -> None:
    """
    Generates a steering file.

    :param output_file: The output filename where the steering file will be save.
    :type output_file: Path | str
    :param processors: The processors list for which the steering file will be generated.
    :type processors: list[type[Processor] | Processor], type[Processor], Processor
    :param database_conf: The database configuration dictionary
    :type database_conf: dict, Optional
    :param db_engine: A string representing the DB engine to be used. Possible values are: *sqlite*, *postgresql*
        and *mysql*.
    :type: str
    """
    if isinstance(output_file, str):
        output_file = Path(output_file)

    doc = _new_toml_doc()
    doc = _add_db_configuration(database_conf, db_engine=db_engine, doc=doc)
    doc = _add_processor_parameters_to_toml_doc(processors, doc)
    doc = _add_user_interface_configuration(doc)

    with open(output_file, 'w') as fp:
        tomlkit.dump(doc, fp)


def _new_toml_doc() -> TOMLDocument:
    doc = document()
    doc.add(comment(f'MAFw steering file generated on {datetime.datetime.now()}'))
    doc.add(nl())
    doc.add(
        comment('uncomment the line below and insert the processors you want to run from the available processor list')
    )
    doc.add(comment('processors_to_run = []'))
    doc.add(nl())
    doc.add(comment('customise the name of the analysis'))
    doc.add('analysis_name', String.from_raw('mafw analysis', StringType.SLB))
    doc.add('analysis_description', String.from_raw('Summing up numbers', StringType.MLB))
    return doc


def _add_db_configuration(
    database_conf: dict[str, Any] | None = None, db_engine: str = 'sqlite', doc: TOMLDocument | None = None
) -> TOMLDocument:
    """Add the DB configuration to the TOML document

    The expected structure of the database_conf dictionary is one of these two:

    .. code-block:: python

        option1 = {
            'DBConfiguration': {
                'URL': 'sqlite:///:memory:',
                'pragmas': {
                    'journal_mode': 'wal',
                    'cache_size': -64000,
                    'foreign_keys': 1,
                    'synchronous': 0,
                },
            }
        }

        option2 = {
            'URL': 'sqlite:///:memory:',
            'pragmas': {
                'journal_mode': 'wal',
                'cache_size': -64000,
                'foreign_keys': 1,
                'synchronous': 0,
            },
        }

    We will always convert the option1 in option2.

    :param database_conf: A dictionary with the database configuration. See comments above. If None, then the default
        is used.
    :type database_conf: dict
    :param db_engine: The database engine. It is used only in case the provided database configuration is invalid to
        retrieve the default configuration. Defaults to sqlite.
    :type db_engine: str, Optional
    :param doc: The TOML document to add the DB configuration. If None, one will be created.
    :type doc: TOMLDocument, Optional
    :return: The modified document.
    :rtype: TOMLDocument
    :raises UnknownDBEngine: if the `database_conf` is invalid and the db_engine is not yet implemented.
    """
    if doc is None:
        doc = _new_toml_doc()

    if database_conf is None:
        if db_engine in default_conf:
            database_conf = default_conf[db_engine]
        else:
            log.critical('The provided db_engine (%s) is not yet implemented', db_engine)
            raise mafw.mafw_errors.UnknownDBEngine(f'DB engine ({db_engine} not implemented')

    is_conf_valid = True
    if 'DBConfiguration' in database_conf:
        # it should be option 1. let's check if there is the URL that is required.
        if 'URL' not in database_conf['DBConfiguration']:
            # no URL
            is_conf_valid = False
        else:
            database_conf = cast(dict[str, Any], database_conf['DBConfiguration'])
    else:
        # option 2
        if 'URL' not in database_conf:
            # no URL
            is_conf_valid = False

    if not is_conf_valid:
        log.error('The provided database configuration is invalid. Adding default configuration')
        if db_engine not in default_conf:
            log.critical('The provided db_engine (%s) is not yet implemented', db_engine)
            raise mafw.mafw_errors.UnknownDBEngine(f'DB engine ({db_engine} not implemented')
        database_conf = default_conf[db_engine]

    db_table = table()
    for key, value in database_conf.items():
        db_table[key] = value
        if key == 'URL':
            db_table[key].comment(
                'Change the protocol depending on the DB type. Update this file to the path of your DB.'
            )
        if key == 'pragmas':
            db_table[key].comment('Leave these default values, unless you know what you are doing!')

    doc.add('DBConfiguration', db_table)
    doc.add(nl())

    return doc


def _add_processor_parameters_to_toml_doc(
    processors: list[type[Processor] | Processor] | type[Processor] | Processor, doc: TOMLDocument | None = None
) -> TOMLDocument:
    if not isinstance(processors, list):
        processors = [processors]

    if not processor_validator(processors):
        raise TypeError('Only processors instances and classes can be accepted')

    if doc is None:
        doc = _new_toml_doc()

    # add an array with all available processors
    doc.add('available_processors', item([p.__name__ if not isinstance(p, Processor) else p.name for p in processors]))
    doc.add(nl())

    # loop over processors
    for p in processors:
        if not isinstance(p, Processor):
            # p is a class not, an instance. so let's create an instance of p
            p = p()

        # create a table for the current processor
        p_table = table()

        # add the first line of the class documentation
        if p.__doc__:
            lines = p.__doc__.splitlines()
            for line in lines:
                line = line.strip()
                if line:
                    p_table.comment(line)
                    break
        # add all parameters to the table, including the help_doc as a comment
        param: PassiveParameter[Any]
        for name, param in p.get_parameters().items():
            p_table[name] = param.value
            if param.doc and type(param.value) is not bool:
                # it looks like due to a strange feature of tomlkit, we cannot add
                # comments to a boolean item. See this issue: https://github.com/python-poetry/tomlkit/issues/390
                p_table[name].comment(param.doc)

        # add the table to the doc and a new line before going to the next item.
        doc.add(p.name, p_table)
        doc.add(nl())

    return doc


def processor_validator(processors: list[type[Processor] | Processor]) -> bool:
    """
    Validates that all items in the list are valid processor instances or classes.

    :param processors: The list of items to be validated.
    :type processors: list[type[Processor] | Processor]
    :return: True if all items are valid.
    :rtype: bool
    """
    return all([isinstance(p, (Processor, type(Processor))) for p in processors])


def dump_processor_parameters_to_toml(
    processors: list[type[Processor] | Processor] | type[Processor] | Processor, output_file: Path | str
) -> None:
    """
    Dumps a toml file with processor parameters.

    This helper function can be used when the parameters of one or many processors have to be dumped to a TOML file.
    For each Processor in the `processors` a table in the TOML file will be added with their parameters is the shape of
    parameter name = value.

    It must be noted that `processors` can be:

        - a list of processor classes (list[type[Processor]])
        - a list of processor instances (list[Processor]])
        - one single processor class (type[Processor])
        - one single processor instance (Processor)

    What value of the parameters will be dumped?
    --------------------------------------------

    Good question, have a look at this :ref:`explanation <parameter_dump>`.

    :param processors: One or more processors for which the parameters should be dumped.
    :type processors: list[type[Processor | Processor]] | type[Processor] | Processor
    :param output_file: The name of the output file for the dump.
    :type output_file: Path | str
    :raise KeyAlreadyPresent: if an attempt to add twice, the same processor is made.
    :raise TypeError: if the list contains items different from Processor classes and instances.
    """

    doc = _add_processor_parameters_to_toml_doc(processors)

    with open(output_file, 'w') as fp:
        tomlkit.dump(doc, fp)


def _add_user_interface_configuration(doc: TOMLDocument | None = None) -> TOMLDocument:
    if doc is None:
        doc = _new_toml_doc()

    ui_table = table()
    ui_table.comment('Specify UI options')
    ui_table['interface'] = 'rich'
    ui_table['interface'].comment('Default "rich", backup "console"')
    doc.add('UserInterface', ui_table)

    return doc


def load_steering_file(steering_file: Path | str, validate: bool = True) -> dict[str, Any]:
    """
    Load a steering file for the execution framework.


    :param steering_file: The path to the steering file.
    :type steering_file: Path, str
    :param validate: A flag to validate the content. Defaults to True.
    :type validate: bool, Optional
    :return: The configuration dictionary.
    :rtype: dict
    :raise FileNotFound: if steering_file does not exist.
    """
    doc = TOMLFile(steering_file).read()

    if validate:
        required_fields = ['processors_to_run', 'UserInterface']
        for field in required_fields:
            if field not in doc.value:
                log.error('Missing section %s in %s' % (field, str(steering_file)))
                raise InvalidSteeringFile(f'Missing {field} in {str(steering_file)}')
        for processor in doc['processors_to_run']:  # type: ignore[union-attr]
            if processor not in doc.value:
                log.error('Missing section %s in %s' % (processor, str(steering_file)))
                raise InvalidSteeringFile(f'Missing {processor} in {str(steering_file)}')

    return doc.value
